"""
Aggregates parser engines together
"""

from . import lin_pcap
from . import win_pcap
from . import pcap_util


def pcap2json(fn, argsj={}):
    """
    argsj: argument dict
    """

    # TODO: add Windows engine back in

    parser = argsj.get("parser", "auto")
    if parser == "auto":
        parser = pcap_util.guess_parser(fn)
        # print("Guess parser: %s" % parser)
        argsj["parser"] = parser

    cls = {
        "lin-pcap": lin_pcap.Gen,
        "lin-pcapng": lin_pcap.Gen,
        "win-pcap": win_pcap.Gen,
        "win-pcapng": win_pcap.Gen,
    }[parser]
    gen = cls(fn, argsj)

    # k,v generator
    return gen.run()


def jgen2j(jgen):
    # Convert generator into static JSON
    j = {}
    for k, v in jgen:
        # Convert nested generator to list
        if k == "data":
            v = list(v)
        j[k] = v
    return j



import binascii

FTDI_DEVICE_OUT_REQTYPE = 64
FTDI_DEVICE_IN_REQTYPE = 192

req_s2i = {
    "RESET": 0,
    "SET_MODEM_CTRL": 1,
    "SET_FLOW_CTRL": 2,
    "SET_BAUDRATE": 3,
    "SET_DATA": 4,
    "POLL_MODEM_STATUS": 0x05,
    "SET_EVENT_CHAR": 0x06,
    "SET_ERROR_CHAR": 0x07,
    "SET_LATENCY_TIMER": 0x09,
    "GET_LATENCY_TIMER": 0x0A,
    "SET_BITMODE": 0x0B,
    "READ_PINS": 0x0C,
    "READ_EEPROM": 0x90,
    "WRITE_EEPROM": 0x91,
    "ERASE_EEPROM": 0x92,
}
req_i2s = dict([(v, k) for k, v in req_s2i.items()])

INTERFACE_ANY = 0
INTERFACE_A = 1
INTERFACE_B = 2
INTERFACE_C = 3
INTERFACE_D = 4


def interface_i2str(i):
    if i == INTERFACE_A:
        return "A"
    elif i == INTERFACE_B:
        return "B"
    elif i == INTERFACE_C:
        return "C"
    elif i == INTERFACE_D:
        return "D"
    assert 0


def flags2dict(s2i, vali):
    ret = {}
    for k, v in s2i.items():
        ret[k] = bool(vali & v)
    return ret


# is a packet json output format but a serial input format
class FT2232CParser(object):
    def __init__(self, args):
        self.ascii = args.ascii
        self.jo = []

    def next_json(self, j, prefix=None):
        self.jo.append(j)

    def header(self):
        # comment("Generated by usbrply-serial FT2232C")
        pass

    def footer(self):
        pass

    def handleControlRead(self, d):
        print(d)
        if d['bRequestType'] != FTDI_DEVICE_IN_REQTYPE:
            return
        request = req_i2s[d['bRequest']]
        print("bRequest", request)
        if request == "POLL_MODEM_STATUS":
            assert d['wLength'] == 2
            buff = bytearray(binascii.unhexlify(d['data']))
            assert buff[1] & 0x0F == 0
            j = {
                # buff1
                # Clear to send
                'CTS': bool(buff[1] & 0x10),
                # Data set ready
                'DTS': bool(buff[1] & 0x20),
                # Ring indicator
                'RI': bool(buff[1] & 0x40),
                # Receive line signal detect
                'RLSD': bool(buff[1] & 0x80),
                # buff0
                # Data ready
                'DR': bool(buff[0] & 0x01),
                # Overrun error
                'OE': bool(buff[0] & 0x02),
                # Parity error
                'PE': bool(buff[0] & 0x04),
                # Framing error
                'FE': bool(buff[0] & 0x08),
                # Break interrupt
                'BI': bool(buff[0] & 0x10),
                # Transmitter holding register
                'THRE': bool(buff[0] & 0x20),
                # Transmitter empty
                'TEMT': bool(buff[0] & 0x40),
                # Error in RCVR FIFO
                'ERR': bool(buff[0] & 0x80),
            }
        else:
            j = {}
            j["type"] = request
            print("%s: FIXME" % (request, ))

        j["type"] = request
        j['rw'] = 'r'
        j['interface'] = interface_i2str(d["wIndex"] & 0xFF)
        self.next_json(j)

    def handleControlWrite(self, d):
        """
        ('bRequest', 'SET_EVENT_CHAR')
        SET_EVENT_CHAR: FIXME
        {'wValue': 0, 'data': '', 'bRequest': 7, 'packn': (301, 302), 'type': 'controlWrite', 'bRequestType': 64, 'wIndex': 1}
        ('bRequest', 'SET_ERROR_CHAR')
        SET_ERROR_CHAR: FIXME
        {'wValue': 1, 'data': '', 'bRequest': 9, 'packn': (303, 304), 'type': 'controlWrite', 'bRequestType': 64, 'wIndex': 1}
        ('bRequest', 'SET_LATENCY_TIMER')
        SET_LATENCY_TIMER: FIXME
        {'wValue': 0, 'data': '', 'bRequest': 2, 'packn': (305, 306), 'type': 'controlWrite', 'bRequestType': 64, 'wIndex': 257}
        ('bRequest', 'SET_FLOW_CTRL')
        {'wValue': 0, 'data': '', 'bRequest': 11, 'packn': (309, 310), 'type': 'controlWrite', 'bRequestType': 64, 'wIndex': 1}
        ('bRequest', 'SET_BITMODE')
        SET_BITMODE: FIXME
        {'wValue': 512, 'data': '', 'bRequest': 11, 'packn': (311, 312), 'type': 'controlWrite', 'bRequestType': 64, 'wIndex': 1}
        ('bRequest', 'SET_BITMODE')
        SET_BITMODE: FIXME
        {'wValue': 2, 'data': '', 'bRequest': 9, 'packn': (621, 622), 'type': 'controlWrite', 'bRequestType': 64, 'wIndex': 1}
        ('bRequest', 'SET_LATENCY_TIMER')
        SET_LATENCY_TIMER: FIXME
        """
        print(d)
        if d['bRequestType'] != FTDI_DEVICE_OUT_REQTYPE:
            return
        request = req_i2s[d['bRequest']]
        print("bRequest", request)

        def SET_FLOW_CTRL(d):
            flag_s2i = {
                "DISABLE_FLOW_CTRL": 0x0,
                "RTS_CTS_HS": 0x01,
                "DTR_DSR_HS": 0x02,
                "XON_XOFF_HS": 0x04,
            }
            return flags2dict(flag_s2i, d["wIndex"] >> 8)

        def SET_DATA(d):
            parity = {
                0: "NONE",
                1: "ODD",
                2: "EVEN",
                3: "MARK",
                4: "SPACE",
            }[(d['wValue'] >> 8) & 0x7]

            stopbits = {
                0: "1",
                1: "15",
                2: "2",
            }[(d['wValue'] >> 11) & 0x3]

            breakon = {
                0: "OFF",
                1: "ON",
            }[(d['wValue'] >> 14) & 0x1]

            j = {
                'parity': parity,
                'stopbits': stopbits,
                'breakon': breakon,
            }

            return j

        def SET_EVENT_CHAR(d):
            """Set the special event character"""
            j = {
                "char": d['wValue'] & 0xFF,
                "enable": bool(d['wValue'] & 0x100),
            }
            return j

        def SET_ERROR_CHAR(d):
            """Set error character"""
            j = {
                "char": d['wValue'] & 0xFF,
                "enable": bool(d['wValue'] & 0x100),
            }
            return j

        def SET_LATENCY_TIMER(d):
            """keeps data in the internal buffer if the buffer is not full yet"""
            latency = d['wValue']
            assert 1 <= latency <= 255
            j = {
                'latency': latency,
            }
            return j

        def SET_BITMODE(d):
            """
            FT_SetBitMode - mode = 0, mask = 0 - Reset the MPSSE controller. Perform a general reset on
            the MPSSE, not the port itself
            
            FT_SetBitMode - mode = 2, mask = 0 - Enable the MPSSE controller. Pin directions are set later
            through the MPSSE commands.
            """
            flag_s2i = {
                "RESET": 0x0,
                "BITBANG": 0x01,
                "MPSSE": 0x02,
                "SYNCBB": 0x04,
                "MCU": 0x08,
                "OPTO": 0x10,
                "CBUS": 0x20,
                "SYNCFF": 0x40,
                "FT1284": 0x80,
            }
            j = flags2dict(flag_s2i, d["wValue"] >> 8)
            j['bitmask'] = d["wValue"] & 0xFF
            return j

        def DEFAULT(d):
            j = {}
            print("%s: FIXME" % (request, ))
            return j

        j = {
            "SET_FLOW_CTRL": SET_FLOW_CTRL,
            "SET_DATA": SET_DATA,
            "SET_EVENT_CHAR": SET_EVENT_CHAR,
            "SET_ERROR_CHAR": SET_ERROR_CHAR,
            "SET_LATENCY_TIMER": SET_LATENCY_TIMER,
            "SET_BITMODE": SET_BITMODE,
        }.get(request, DEFAULT)(d)

        j["type"] = request
        j['rw'] = 'w'
        j['interface'] = interface_i2str(d["wIndex"] & 0xFF)
        self.next_json(j)

    def handleBulkWrite(self, d):
        # print(d)
        # json encodes in hex
        # protocol itself encodes in hex
        # data = binascii.unhexlify(d["data"])
        # print(len(data))

        interface = {
            0x02: 0,
            0x04: 1,
        }[d["endp"]]

        self.next_json({
            "type": "write",
            "interface": interface,
            "data": d["data"],
        })

    def handleBulkRead(self, d):
        assert len(d["data"]) % 2 == 0
        # json encodes in hex
        # protocol itself encodes in hex
        data = binascii.unhexlify(d["data"])
        # print(d)

        interface = {
            0x81: 0,
            0x83: 1,
        }[d["endp"]]

        prefix = data[0:2]
        data = data[2:]
        # meh lots of these and not sure what they mean
        # should look into these but just ignore for now
        # assert prefix == "\x42\x60" or prefix == "\x32\x60" or prefix == "\x32\x00", d

        if len(data):
            self.next_json({
                "type": "read",
                "interface": interface,
                "data": binascii.hexlify(data),
                "prefix": prefix,
            })

    def run(self, j):
        self.header()

        for di, d in enumerate(j["data"]):
            if 0 and di > 500:
                print("debug break")
                break
            if d["type"] == "bulkWrite":
                self.handleBulkWrite(d)
            elif d["type"] == "bulkRead":
                self.handleBulkRead(d)
            elif d["type"] == "controlRead":
                self.handleControlRead(d)
            elif d["type"] == "controlWrite":
                self.handleControlWrite(d)

        self.footer()
        j = {
            "data": self.jo,
        }
        return j



from .usb import req2s
import binascii
import struct
from .util import hexdump

# FIXME: this is super hacky right now


def default_arg(argsj, k, default):
    val = argsj.get(k)
    if val is None:
        return default
    else:
        return val


class DeviceFilter(object):
    def __init__(self, argsj, verbose=False):
        self.verbose = verbose
        self.entries = 0
        self.drops = 0

        self.arg_device = default_arg(argsj, "device", None)

    def should_filter(self, data):
        device = data.get('device')
        # Filter:
        # Devices not matching target
        # When device filtering not used
        return device is not None and device == self.keep_device

    def gen_data(self, datas):
        for data in datas:
            self.entries += 1
            if self.should_filter(data):
                if self.verbose:
                    print("DeviceFilter drop %s (%s %s %s)" %
                          (data['type'],
                           req2s(data["bRequestType"], data["bRequest"]),
                           data["bRequestType"], data["bRequest"]))
                self.drops += 1
                continue
            yield data
        yield {
            "type":
            "comment",
            "v":
            "DeviceFilter: dropped %s / %s entries" %
            (self.drops, self.entries)
        }

    def run(self, jgen):
        for k, v in jgen:
            if k == "data":
                yield k, self.gen_data(v)
            else:
                yield k, v



from .printer import JSONPrinter
from .pyprinter import LibusbPyPrinter


def run(ofmt, j, argsj={}):
    cls = {
        "json": JSONPrinter,
        # "libusb-c": LibusbCPrinter,
        "libusb-py": LibusbPyPrinter,
        # "linux": LinuxPrinter,
    }[ofmt]

    printer = cls(argsj)
    printer.run(j)



import binascii
from usbrply.printer import Printer, indented, indent_inc, indent_dec
from usbrply.pyprinter import bytes2AnonArray


class TextFT2232CPrinter(object):
    def __init__(self, args):
        self.ascii = args.ascii

    def next_json(self, j, prefix=None):
        if j['type'] == 'read':
            interface = j['interface']
            data = j['data']
            n = len(data) / 2
            prefix = j['prefix']
            if self.ascii:
                indented("%u r %s: %s" %
                         (interface, n, binascii.unhexlify(data)))
            else:
                indented("%u r %s: 0x%s" % (interface, n, data))
                indented("  %s" % (binascii.hexlify(prefix, )))
        elif j['type'] == 'write':
            interface = j['interface']
            data = j['data']
            n = len(data) / 2
            if self.ascii:
                indented("%u w %s: %s" %
                         (interface, n, binascii.unhexlify(data)))
            else:
                indented("%u w %s: 0x%s" % (interface, n, data))
        elif j['type'] == 'SET_DATA':
            indented(
                "i%s SET_DATA: parity %s, stop bits %s, break %s" %
                (j['interface'], j['parity'], j['stopbits'], j['breakon']))
        else:
            indented("i%s %s: %s" % (j['interface'], j['type'], j))

    def comment(self, s):
        indented('%s' % (s, ))

    def header(self):
        self.comment("Generated by usbrply-serial FT2232C")

    def footer(self):
        pass

    def run(self, j):
        print("")
        print("")
        print("")

        self.header()

        for d in j["data"]:
            self.next_json(d)

        self.footer()


class JSONSPrinter(object):
    def run(self, j):
        print(json.dumps(j, sort_keys=True, indent=4, separators=(',', ': ')))


class PythonFT2232CPrinter(object):
    def __init__(self, args):
        self.ascii = args.ascii
        self.wrapper = args.wrapper

    def next_json(self, j, prefix=None):
        if j['type'] == 'read':
            interface = j['interface']
            data = binascii.unhexlify(j["data"])
            indented("buff = ser%u.read(%u)" % (interface, len(data)))
            indented("validate_read(%s, buff, \"%s\")" %
                     (bytes2AnonArray(data), "packet"))
        elif j['type'] == 'write':
            data_str = bytes2AnonArray(binascii.unhexlify(j["data"]))
            interface = j['interface']
            indented("ser%u.write(%s)" % (interface, data_str))
        else:
            # assert 0
            print('# next_json: %s' % (j['type'], ))

    def comment(self, s):
        indented('# %s' % (s, ))

    def header(self):
        if not self.wrapper:
            return
        # self.comment("Generated by usbrply-serial FT2232C")
        print('''
#!/usr/bin/env python3
import binascii
import time
import serial

def validate_read(expected, actual, msg):
    if expected != actual:
        print('Failed %s' % msg)
        print('  Expected; %s' % binascii.hexlify(expected,))
        print('  Actual:   %s' % binascii.hexlify(actual,))
        #raise Exception('failed validate: %s' % msg)

''')
        print('def replay(ser0):')
        indent_inc()

    def footer(self):
        indent_dec()
        if not self.wrapper:
            return
        print('''
if __name__ == "__main__":
    import argparse 
    
    parser = argparse.ArgumentParser(description='Replay captured USB-serial packets')
    parser.add_argument('-p',
                        "--port",
                        default="/dev/ttyUSB0",
                        help="")
    args = parser.parse_args()

    serial0 = serial.Serial(
        args.port,
        baudrate=115200,
        bytesize=serial.EIGHTBITS,
        parity=serial.PARITY_NONE,
        stopbits=serial.STOPBITS_ONE,
        rtscts=False,
        dsrdtr=False,
        xonxoff=False,
        timeout=0,
        writeTimeout=None)
    replay(serial0)

''')

    def run(self, j):
        self.header()

        for d in j["data"]:
            self.next_json(d)

        self.footer()



#!/usr/bin/env python

from .usb import *

# High performance C pcap librar
# Python 2 only, no longer maintained
try:
    import pcap
except ImportError:
    pcap = None

# Slow but reliable pure Python
try:
    import pcapng
except ImportError:
    pcapng = None

import sys
"""
Quick hack to detect packet format
I don't think the API i'm using 
"""


def guess_linux(buff):
    """
    linux heuristics
    0x8: one of SCE
    0x1C:0x1F (urb status): 0 => success, almost always
        windows: 
    """
    if len(buff) < 0x30:
        return False
    return sum(buff[0x1C:0x20]) == 0


def guess_windows(buff):
    """
    windows heuristics
    0xA:0xD (error code): 0 => success, almost always
        linux: endpoint, device, bus id. Unlikely to be 0
    0x10 (IRP information): either 0 or 1
    """
    if len(buff) < 0x24:
        return False
    return sum(buff[0x0A:0x0E]) == 0


class PcapParser(object):
    def __init__(self, fn, use_pcapng=None):
        self.fn = fn

        # Select library
        self.use_pcapng = use_pcapng
        if self.use_pcapng is None:
            # User higher performance library if available
            self.use_pcapng = False if pcap else True
        # self.pcapng = "pcapng" in argsj["parser"]
        if self.use_pcapng:
            assert pcapng, "pcapng library requested but no pcapng library"
        else:
            assert pcap, "pcap library requested but no pcap library"

        # Initialize library
        if self.use_pcapng:
            self.fp = open(fn, 'rb')
            self.scanner = pcapng.FileScanner(self.fp)
            self.scanner_iter = self.scanner.__iter__()
        else:
            self.pcap = pcap.pcapObject()
            self.pcap.open_offline(fn)

    def next(self, loop_cb):
        """return True if there was data and might be more, False if nothing was processed"""
        if self.use_pcapng:
            while True:
                try:
                    block = next(self.scanner_iter)
                except StopIteration:
                    return False

                if not isinstance(block, pcapng.blocks.EnhancedPacket):
                    continue
                loop_cb(block.captured_len, block.packet_data, block.timestamp)
                return True
        else:
            got = [False]

            # return code isn't given to indicate end
            def my_loop_cb(*args, **kwargs):
                got[0] = True
                loop_cb(*args, **kwargs)

            self.pcap.loop(1, my_loop_cb)
            return got[0]


def load_pcap(fn, loop_cb, lim=float('inf'), use_pcapng=None):
    parser = PcapParser(fn, use_pcapng=use_pcapng)
    i = 0
    while parser.next(loop_cb):
        i += 1
        if i >= lim:
            break


def guess_parser(fn):
    windows = [0]
    linux = [0]

    def loop_cb_guess(caplen, packet, ts):
        packet = bytearray(packet)
        if guess_linux(packet):
            linux[0] += 1
        if guess_windows(packet):
            windows[0] += 1

    parser = PcapParser(fn)
    i = 0
    while parser.next(loop_cb_guess):
        i += 1
        if i >= 3:
            break

    if windows[0]:
        assert linux[0] == 0
        if parser.use_pcapng:
            return "win-pcapng"
        else:
            return "win-pcap"
    if linux[0]:
        assert windows[0] == 0
        if parser.use_pcapng:
            return "lin-pcapng"
        else:
            return "lin-pcap"
    assert 0, "failed to identify packet format"



#!/usr/bin/env python
'''
Windows is quirky
http://desowin.org/usbpcap/capture_limitations.html

Control packets
-What are the status packets used for?
-Control write has the request on the response packet

Bulk packets
-Only one packet for both request and response
'''

from .usb import *
from .util import hexdump, tostr
from .com_pcap import PcapGen

import sys
import binascii
import struct
from collections import namedtuple

# Windows transfer stages
# Usb function: URB_FUNCTION_VENDOR_DEVICE (0x17)
XFER_SETUP = 0
# Usb function: URB_FUNCTION_CONTROL_TRANSFER (0x08)
XFER_DATA = 1
# Usb function: URB_FUNCTION_CONTROL_TRANSFER (0x08)
XFER_STATUS = 2

USBD_STATUS_SUCCESS = 0

# https://msdn.microsoft.com/en-us/library/windows/hardware/ff540409(v=vs.85).aspx
# https://github.com/wine-mirror/wine/blob/master/include/ddk/usb.h
URB_FUNCTION_CONTROL_TRANSFER = 0x08
URB_FUNCTION_VENDOR_DEVICE = 0x17

func_i2s = {
    0x0000: "SELECT_CONFIGURATION",
    0x0001: "SELECT_INTERFACE",
    0x0002: "ABORT_PIPE",
    0x0003: "TAKE_FRAME_LENGTH_CONTROL",
    0x0004: "RELEASE_FRAME_LENGTH_CONTROL",
    0x0005: "GET_FRAME_LENGTH",
    0x0006: "SET_FRAME_LENGTH",
    0x0007: "GET_CURRENT_FRAME_NUMBER",
    0x0008: "CONTROL_TRANSFER",
    0x0009: "BULK_OR_INTERRUPT_TRANSFER",
    0x000A: "ISOCH_TRANSFER",
    0x000B: "GET_DESCRIPTOR_FROM_DEVICE",
    0x000C: "SET_DESCRIPTOR_TO_DEVICE",
    0x000D: "SET_FEATURE_TO_DEVICE",
    0x000E: "SET_FEATURE_TO_INTERFACE",
    0x000F: "SET_FEATURE_TO_ENDPOINT",
    0x0010: "CLEAR_FEATURE_TO_DEVICE",
    0x0011: "CLEAR_FEATURE_TO_INTERFACE",
    0x0012: "CLEAR_FEATURE_TO_ENDPOINT",
    0x0013: "GET_STATUS_FROM_DEVICE",
    0x0014: "GET_STATUS_FROM_INTERFACE",
    0x0015: "GET_STATUS_FROM_ENDPOINT",
    0x0016: "RESERVED_0X0016",
    0x0017: "VENDOR_DEVICE",
    0x0018: "VENDOR_INTERFACE",
    0x0019: "VENDOR_ENDPOINT",
    0x001A: "CLASS_DEVICE",
    0x001B: "CLASS_INTERFACE",
    0x001C: "CLASS_ENDPOINT",
    0x001D: "RESERVE_0X001D",
    0x001E: "SYNC_RESET_PIPE_AND_CLEAR_STALL",
    0x001F: "CLASS_OTHER",
    0x0020: "VENDOR_OTHER",
    0x0021: "GET_STATUS_FROM_OTHER",
    0x0022: "CLEAR_FEATURE_TO_OTHER",
    0x0023: "SET_FEATURE_TO_OTHER",
    0x0024: "GET_DESCRIPTOR_FROM_ENDPOINT",
    0x0025: "SET_DESCRIPTOR_TO_ENDPOINT",
    0x0026: "GET_CONFIGURATION",
    0x0027: "GET_INTERFACE",
    0x0028: "GET_DESCRIPTOR_FROM_INTERFACE",
    0x0029: "SET_DESCRIPTOR_TO_INTERFACE",
    0x002A: "GET_MS_FEATURE_DESCRIPTOR",
    0x002B: "RESERVE_0X002B",
    0x002C: "RESERVE_0X002C",
    0x002D: "RESERVE_0X002D",
    0x002E: "RESERVE_0X002E",
    0x002F: "RESERVE_0X002F",
    0x0030: "SYNC_RESET_PIPE",
    0x0031: "SYNC_CLEAR_STALL",
}


def func_str(func):
    return func_i2s.get(func, "0x%04X" % func)


'''
struct usb_ctrlrequest {
    __u8 bRequestType;
    __u8 bRequest;
    __le16 wValue;
    __le16 wIndex;
    __le16 wLength;
} __attribute__ ((packed));
'''
usb_ctrlrequest_nt = namedtuple('usb_ctrlrequest_win', (
    'bRequestType',
    'bRequest',
    'wValue',
    'wIndex',
    'wLength',
))
usb_ctrlrequest_fmt = '<BBHHH'
usb_ctrlrequest_sz = struct.calcsize(usb_ctrlrequest_fmt)


def usb_ctrlrequest(s):
    return usb_ctrlrequest_nt(
        *struct.unpack(usb_ctrlrequest_fmt, bytearray(s)))


# https://stackoverflow.com/questions/19110075/what-is-the-difference-between-pdo-and-fdo-in-windows-device-drivers
# PDO = Physical Device Object
# FDO = Functional Device Object
# Submit
INFO_FDO2PDO = 0
# Complete
INFO_PDO2FDO = 1


def irp_info_str(irp_info):
    if irp_info & 1:
        return "PDO2FDO (0x%02X)" % irp_info
    else:
        return "FDO2PDO (0x%02X)" % irp_info


'''
typedef struct {
    uint64_t id
    uint8_t type
    uint8_t transfer_type
    uint8_t endpoint
    uint8_t device
    uint16_t bus_id
    uint8_t setup_request
    uint8_t data
    uint64_t sec
    uint32_t usec
    uint32_t status
    uint32_t length
    uint32_t data_length
    //These form the URB setup for control transfers and are techincally part of the URB
    //uint8_t pad[24]
} __attribute__((packed)) usb_urb_t
'''
'''
IRP: I/O request packet
https://msdn.microsoft.com/en-us/library/windows/hardware/ff550694(v=vs.85).aspx

also useful info
https://www.wireshark.org/docs/dfref/u/usb.html
'''
# Header
# Packet may have additional data
usb_urb_win_nt = namedtuple(
    'usb_urb_win',
    (
        # Length of entire packet entry including htis header and additional pkt_len data
        'pcap_hdr_len',
        # IRP ID
        # buffer ID or something like that
        # it is not a unique packet ID
        # but can be used to match up submit and response
        'id',
        # IRP_USBD_STATUS
        'irp_status',
        # USB Function
        'usb_func',
        # IRP Information
        # Ex: Direction: PDO => FDO
        'irp_info',
        # USB port
        # Ex: 3
        'bus_id',
        # USB device on that port
        # Ex: 16
        'device',
        # Which endpoint on that bus
        # Ex: 0x80 (0 in)
        'endpoint',
        # Ex: URB_CONTROL
        'transfer_type',
        # Length of data beyond header
        'data_length',
    ))
usb_urb_win_fmt = (
    '<'
    'H'  # pcap_hdr_len
    'Q'  # irp_id
    'I'  # irp_status
    'H'  # usb_func
    'B'  # irp_info
    'H'  # bus_id
    'H'  # device
    'B'  # endpoint
    'B'  # transfer_type
    'I'  # data_length
)

#print 'WARNING: experimental windows mode activated'
usb_urb_nt = usb_urb_win_nt
usb_urb_fmt = usb_urb_win_fmt

usb_urb_sz = struct.calcsize(usb_urb_fmt)


def usb_urb(s):
    return usb_urb_nt(*struct.unpack(usb_urb_fmt, bytes(s)))


# When we get an IN request we may process packets in between
class PendingRX:
    def __init__(self):
        # Unprocessed packet bytes
        self.raw = None
        #usb_urb_t m_urb
        self.m_urb = None
        #usb_ctrlrequest m_ctrl
        # Only applies to control requests
        self.m_ctrl = None
        self.packet_number = 0

        # uint8_t *m_data_out
        self.m_data_out = None


class payload_bytes_type_t:
    def __init__(self):
        self.req_in = 0
        self.req_in_last = None
        self.in_ = 0
        self.in_last = None

        self.req_out = 0
        self.req_out_last = None
        self.out = 0
        self.out_last = None


class payload_bytes_t:
    def __init__(self):
        self.ctrl = payload_bytes_type_t()
        self.bulk = payload_bytes_type_t()


g_payload_bytes = payload_bytes_t()


def update_delta(pb):
    pb.req_in_last = pb.req_in
    pb.in_last = pb.in_

    pb.req_out_last = pb.req_out
    pb.out_last = pb.out


def bytes2Hex(bytes_data):
    return tostr(binascii.hexlify(bytes_data))


def deviceStr():
    # return "dev.udev"
    return "udev"


def print_urb(urb):
    # unique per transaction, not packet
    print("URB id: %s" % (urb_id_str(urb.id)))
    print(" pcap_hdr_len: %s" % (urb.pcap_hdr_len, ))
    print(" irp_status: %s" % (urb.irp_status, ))
    print(" usb_func: %s" % (func_str(urb.usb_func), ))
    print(" irp_info: %s" % (irp_info_str(urb.irp_info), ))
    print(" bus_id: %s" % (urb.bus_id, ))
    print(" device: %s" % (urb.device, ))
    print(" endpoint: 0x%02X" % (urb.endpoint, ))
    print(" transfer_type: %s" % (urb.transfer_type, ))
    print(" data_length: %s" % (urb.data_length, ))


def urb2json(urb):
    j = dict(urb._asdict())
    #j["ctrlrequest"] = binascii.hexlify(j["ctrlrequest"])
    # j["data"] = binascii.hexlify(j["data"])
    #j["t"] = j['sec'] + j['usec'] / 1e6
    return j


def urb_error(urb):
    return urb.irp_status != USBD_STATUS_SUCCESS


def is_urb_submit(urb):
    return urb.usb_func == URB_FUNCTION_VENDOR_DEVICE


def is_urb_complete(urb):
    return urb.usb_func == URB_FUNCTION_CONTROL_TRANSFER


def urb_id_str(urb_id):
    # return binascii.hexlify(urb_id)
    return '0x%X' % urb_id


class Gen(PcapGen):
    def __init__(self, fn, argsj={}):
        PcapGen.__init__(self, argsj)

        self.arg_fin = fn
        self.cur_packn = 0
        self.rel_pkt = 0

        self.previous_urb_complete_kept = None
        self.pending_complete = {}
        self.errors = 0

    def comment_source(self):
        self.gcomment('Source: Windows pcap (USBPcap)')

    def parser(self):
        return "win-pcap"

    def platform(self):
        return "windows"

    def loop_cb(self, caplen, packet, ts):
        """
        2020-12-22
        Looks like there are two different structures
        Example observed on controlWrite

        Type 1
        Packet1: out FDO -> PDO
        Packet2: out PDO -> FDO

        Type 2
        Packet1: out FDO -> PDO
        Packet2: out FDO -> PDO
        Packet3: status PDO -> FDO

        Not sure exactly what this is (different windows version?)
        But lets make sure to handle both types
        """
        try:
            packet = bytearray(packet)
            self.cur_packn += 1
            #if self.cur_packn >= 871:
            #    self.verbose = True

            if self.cur_packn < self.min_packet or self.cur_packn > self.max_packet:
                # print("# Skipping packet %d" % (self.cur_packn))
                return
            if self.verbose:
                print("")
                print("")
                print("")
                print('WIN PACKET %s' % (self.cur_packn, ))

            if caplen != len(packet):
                self.gwarning("packet %s: malformed, caplen %d != len %d",
                              self.pktn_str(), caplen, len(packet))
                return
            if self.verbose:
                # print('Len: %d' % len(packet))
                hexdump(packet)
                #print(ts)
                print('Pending (%d):' % (len(self.pending_complete), ))
                for k in self.pending_complete:
                    print('  %s' % (urb_id_str(k), ))

            self.printv("Length %u" % (len(packet), ))
            if len(packet) < usb_urb_sz:
                msg = "Packet %s: size %d is not min size %d" % (
                    self.pktn_str(), len(packet), usb_urb_sz)
                self.errors += 1
                if self.arg_halt:
                    hexdump(packet)
                    raise ValueError(msg)
                if self.verbose:
                    print(msg)
                    hexdump(packet)
                return

            # caplen is actual length, len is reported
            self.urb_raw = packet
            self.urb = usb_urb(packet[0:usb_urb_sz])
            dat_cur = packet[usb_urb_sz:]

            self.printv('ID %s, %s post-urb bytes' %
                        (urb_id_str(self.urb.id), len(dat_cur)))

            # Main packet filtering
            # Drop if not specified device
            #print(self.pktn_str(), self.urb.device, args.device)
            if self.arg_device is not None and self.urb.device != self.arg_device:
                self.printv("Drop: device mismatch")
                self.dev_drops += 1
                return
            '''
            # Status package may contain complete
            # But it looks like any
            if self.urb.transfer_type == URB_CONTROL:
                # Control transfer stage
                # 1: data
                # 2: status
                # 'xfer_stage',
                xfer_stage = dat_cur[0]
                """
                #print('xfer_stage: %d' % xfer_stage)
                if xfer_stage == XFER_STATUS:
                    if not self.urb.id in self.pending_complete:
                        if self.urb.transfer_type != URB_INTERRUPT:
                            self.gwarning(
                                "Packet %s missing submit.  URB ID: 0x%016lX" %
                                (self.pktn_str(), self.urb.id))
                    else:
                        self.process_complete(dat_cur)
                """
            '''

            self.rel_pkt += 1

            if self.verbose:
                # print("Header size: %lu" % (usb_urb_sz,))
                print_urb(self.urb)

            if urb_error(self.urb):
                self.errors + 1
                if self.arg_halt:
                    print("oh noes!")
                    sys.exit(1)

            # Complete?
            # May be "out" or "status"
            if self.urb.irp_info & 1 == INFO_PDO2FDO:
                if not self.urb.id in self.pending_complete:
                    if self.urb.transfer_type != URB_INTERRUPT:
                        self.gwarning(
                            "Packet %s missing submit.  URB ID: 0x%016lX" %
                            (self.pktn_str(), self.urb.id))
                else:
                    self.process_complete(dat_cur)
            # Oterhwise submit
            else:
                # Find the matching submit request
                if self.urb.transfer_type == URB_CONTROL:
                    self.processControlSubmit(dat_cur)
                elif self.urb.transfer_type == URB_BULK:
                    self.processBulkSubmit(dat_cur)
                elif self.urb.transfer_type == URB_INTERRUPT:
                    self.processInterruptSubmit(dat_cur)

            assert len(self.pcomments) == 0
            self.submit = None
            self.urb = None
        except:
            print('ERROR: packet %s' % self.pktn_str())
            raise

    def pktn_str(self):
        if self.arg_rel_pkt:
            return self.rel_pkt
        else:
            return self.cur_packn

    def process_complete(self, dat_cur):
        self.printv("process_complete")
        self.submit = self.pending_complete[self.urb.id]
        # Done with it, get rid of it
        del self.pending_complete[self.urb.id]
        self.printv("Matched submit packet %s" % self.submit.packet_number)

        # Discarded?
        if self.submit is None:
            return

        self.packnum()

        if self.previous_urb_complete_kept is not None:
            '''
            For bulk packets this can get tricky
            The intention was mostly for control packets where timing might be more critical
            '''

        self.previous_urb_complete_kept = self.urb

        # Find the matching submit request
        if self.urb.transfer_type == URB_CONTROL:
            if self.submit.m_ctrl.bRequestType & URB_TRANSFER_IN:
                self.processControlCompleteIn(dat_cur)
            else:
                self.processControlCompleteOut(dat_cur)
        elif self.urb.transfer_type == URB_BULK:
            if self.urb.endpoint & URB_TRANSFER_IN:
                g_payload_bytes.bulk.in_ += self.urb.data_length
                self.processBulkCompleteIn(dat_cur)
            else:
                g_payload_bytes.bulk.out += self.urb.data_length
                self.processBulkCompleteOut(dat_cur)
        elif self.urb.transfer_type == URB_INTERRUPT:
            if self.urb.endpoint & URB_TRANSFER_IN:
                self.processInterruptCompleteIn(dat_cur)
            else:
                self.processInterruptCompleteOut(dat_cur)
        else:
            self.verbose and print(
                "WARNING: unknown transfer type %u" % self.urb.transfer_type)

    def processControlSubmit(self, dat_cur):
        pending = PendingRX()
        pending.raw = self.urb_raw
        pending.m_urb = self.urb

        self.printv('Remaining data: %d' % (len(dat_cur)))
        #self.printv('ctrlrequest: %d' % (len(self.urb.ctrlrequest)))
        # Skip xfer_stage
        dat_cur = dat_cur[1:]
        ctrl = usb_ctrlrequest(dat_cur[0:usb_ctrlrequest_sz])
        dat_cur = dat_cur[usb_ctrlrequest_sz:]

        if self.verbose:
            print("Packet %s control submit (control info size %lu)" %
                  (self.pktn_str(), 666))
            print("    bRequestType: %s (0x%02X)" %
                  (request_type2str(ctrl.bRequestType), ctrl.bRequestType))
            #print("    bRequest: %s (0x%02X)" % (request2str(ctrl), ctrl.bRequest))
            print("    wValue: 0x%04X" % (ctrl.wValue))
            print("    wIndex: 0x%04X" % (ctrl.wIndex))
            print("    wLength: 0x%04X" % (ctrl.wLength))

        if (ctrl.bRequestType & URB_TRANSFER_IN) == URB_TRANSFER_IN:
            self.printv("%d: IN" % (self.cur_packn))
        else:
            self.printv("%d: OUT" % (self.cur_packn))
            pending.m_data_out = bytearray(dat_cur)

        pending.m_ctrl = ctrl
        pending.packet_number = self.pktn_str()
        self.pending_complete[self.urb.id] = pending
        self.printv('Added pending control URB %s, len %d' %
                    (urb_id_str(self.urb.id), len(self.pending_complete)))

    def processControlCompleteIn(self, dat_cur):
        packet_numbering = ''
        data_size = 0
        data_str = "None"
        max_payload_sz = self.submit.m_ctrl.wLength

        # Skip xfer_stage
        dat_cur = dat_cur[1:]
        #print 'shorten'

        # Is it legal to have a 0 length control in?
        if self.submit.m_ctrl.wLength:
            data_str = "buff"
            data_size = self.submit.m_ctrl.wLength

        # Verify we actually have enough / expected
        # If exact match don't care
        if len(dat_cur) != max_payload_sz:
            if len(dat_cur) < max_payload_sz:
                self.gcomment("NOTE:: req max %u but got %u" %
                              (max_payload_sz, len(dat_cur)))
            else:
                raise Exception('invalid response')

        self.output_packet({
            'type': 'controlRead',
            'bRequestType': self.submit.m_ctrl.bRequestType,
            'bRequest': self.submit.m_ctrl.bRequest,
            'wValue': self.submit.m_ctrl.wValue,
            'wIndex': self.submit.m_ctrl.wIndex,
            'wLength': self.submit.m_ctrl.wLength,
            'data': bytes2Hex(dat_cur)
        })

        if self.submit.m_ctrl.wLength:
            if self.arg_packet_numbers:
                packet_numbering = "packet %s/%s" % (self.submit.packet_number,
                                                     self.pktn_str())
            else:
                # TODO: consider counting instead of by captured index
                packet_numbering = "packet"

    def processControlCompleteOut(self, dat_cur):
        data_size = 0
        data_str = "None"
        data = None

        #print 'Control out w/ len %d' % len(submit.m_data_out)

        # print "Data out size: %u vs urb size %u" % (submit.m_data_out_size, submit.m_urb.data_length )
        # For some reason the request data is in the reply
        # Skip xfer_type
        data = dat_cur[1:]

        if data:
            data_str = bytes2Hex(data)
            data_size = len(data)

        self.output_packet({
            'type': 'controlWrite',
            'bRequestType': self.submit.m_ctrl.bRequestType,
            'bRequest': self.submit.m_ctrl.bRequest,
            'wValue': self.submit.m_ctrl.wValue,
            'wIndex': self.submit.m_ctrl.wIndex,
            'data': bytes2Hex(data)
        })

    def output_packet(self, j):
        urbj_submit = urb2json(self.submit.m_urb)
        urbj_complete = urb2json(self.urb)
        nsub, ncomplete = self.packnumt()
        j["submit"] = {
            "packn": nsub,
            'urb': urbj_submit,
            # 't': urbj_submit["t"],
        }
        j["complete"] = {
            'packn': ncomplete,
            'urb': urbj_complete,
            # 't': urbj_complete["t"],
        }
        if len(self.pcomments):
            j["comments"] = self.pcomments
        self.jbuff.append(j)
        self.pcomments = []

    def processBulkSubmit(self, dat_cur):
        if self.urb.endpoint & URB_TRANSFER_IN:
            g_payload_bytes.bulk.req_in += self.urb.data_length
        else:
            g_payload_bytes.bulk.req_out += self.urb.data_length

        pending = PendingRX()
        pending.raw = self.urb_raw
        pending.m_urb = self.urb

        self.printv('Remaining data: %d' % (len(dat_cur)))

        # self.printv("Packet %d bulk submit (control info size %lu)" % (self.pktn_str(), 666))

        if self.urb.endpoint & URB_TRANSFER_IN:
            self.printv("%d: IN" % (self.cur_packn))
        else:
            self.printv("%d: OUT" % (self.cur_packn))
            if len(dat_cur) != self.urb.data_length:
                self.gcomment(
                    "WARNING: remaining bytes %d != expected payload out bytes %d"
                    % (len(dat_cur), self.urb.data_length))
                hexdump(dat_cur, "  ")
                raise Exception('See above')
            pending.m_data_out = bytearray(dat_cur)

        pending.packet_number = self.pktn_str()
        self.pending_complete[self.urb.id] = pending
        self.printv('Added pending bulk URB %s' % self.urb.id)

    def processBulkCompleteIn(self, dat_cur):
        # looks like maybe windows doesn't report the request size?
        # think this is always 0
        assert self.submit.m_urb.data_length == 0

        # FIXME: this is a messy conversion artifact from the C code
        # Is it legal to have a 0 length bulk in?
        data_size = 0
        # instead, use the recieved buffer size as a best estimated
        max_payload_sz = len(dat_cur)
        if max_payload_sz:
            data_size = max_payload_sz

        # output below
        self.output_packet({
            'type': 'bulkRead',
            'endp': self.submit.m_urb.endpoint,
            'len': data_size,
            'data': bytes2Hex(dat_cur)
        })

    def processInterruptSubmit(self, dat_cur):
        pending = PendingRX()
        pending.raw = self.urb_raw
        pending.m_urb = self.urb
        pending.packet_number = self.pktn_str()
        self.pending_complete[self.urb.id] = pending
        self.printv('Added pending interrupt URB %s' % self.urb.id)

    def processInterruptCompleteOut(self, dat_cur):
        # looks like maybe windows doesn't report the request size?
        # think this is always 0
        assert self.submit.m_urb.data_length == 0

        # FIXME: this is a messy conversion artifact from the C code
        # Is it legal to have a 0 length bulk in?
        data_size = 0
        # instead, use the recieved buffer size as a best estimate
        max_payload_sz = len(dat_cur)
        if max_payload_sz:
            data_size = max_payload_sz

        self.output_packet({
            'type': 'interruptWrite',
            'endp': self.submit.m_urb.endpoint,
            'len': data_size,
            'data': bytes2Hex(dat_cur)
        })

    def processInterruptCompleteIn(self, dat_cur):
        # looks like maybe windows doesn't report the request size?
        # think this is always 0
        assert self.submit.m_urb.data_length == 0

        # FIXME: this is a messy conversion artifact from the C code
        # Is it legal to have a 0 length bulk in?
        data_size = 0
        # instead, use the recieved buffer size as a best estimate
        max_payload_sz = len(dat_cur)
        if max_payload_sz:
            data_size = max_payload_sz

        # output below
        self.output_packet({
            'type': 'interruptRead',
            'endp': self.submit.m_urb.endpoint,
            'len': data_size,
            'data': bytes2Hex(dat_cur)
        })

    def processBulkCompleteOut(self, dat_cur):
        self.output_packet({
            'type': 'bulkWrite',
            'endp': self.submit.m_urb.endpoint,
            'data': bytes2Hex(self.submit.m_data_out)
        })

    def loop_cb_devmax(self, caplen, packet, ts):
        self.cur_packn += 1
        if self.cur_packn < self.min_packet or self.cur_packn > self.max_packet:
            # print("# Skipping packet %d" % (self.cur_packn))
            return

        if caplen != len(packet):
            print("packet %s: malformed, caplen %d != len %d", self.pktn_str(),
                  caplen, len(packet))
            return
        #if self.verbose:
        #    print('Len: %d' % len(packet))

        # dbg("Length %u" % (len(packet),))
        if len(packet) < usb_urb_sz:
            hexdump(packet)
            raise ValueError("Packet size %d is not min size %d" %
                             (len(packet), usb_urb_sz))

        # caplen is actual length, len is reported
        self.urb_raw = packet
        self.urb = usb_urb(packet[0:usb_urb_sz])
        dat_cur = packet[usb_urb_sz:]

        self.arg_device = max(self.arg_device, self.urb.device)



"""
Parses raw FTDI calls into higher level MPSSE info

If a bad command is detected, the MPSSE returns the value 0xFA, followed by the byte that caused the
bad command.

Use of the bad command detection is the recommended method of determining whether the MPSSE is in
sync with the application program. By sending a bad command on purpose and looking for 0xFA, the
application can determine whether communication with the MPSSE is possible.
"""

import binascii

mpsse_cmd_s2i = {
    # think values < 0x10 are bit bang?

    # 3.3 MSB FIRST
    "MSB_DOUT_BYTES_PVE": 0x10,
    "MSB_11": 0x11,
    "MSB_12": 0x12,
    "MSB_13": 0x13,
    "MSB_20": 0x20,
    "MSB_24": 0x24,
    "MSB_22": 0x22,
    "MSB_26": 0x26,
    "MSB_31": 0x31,
    "MSB_34": 0x34,
    "MSB_33": 0x33,
    "MSB_36": 0x36,

    # 3.4 LSB FIRST
    "LSB_18": 0x18,
    "LSB_19": 0x19,
    "LSB_1A": 0x1A,
    "LSB_1B": 0x1B,
    "LSB_28": 0x28,
    "LSB_2C": 0x2C,
    "LSB_2A": 0x2A,
    "LSB_2E": 0x2E,
    "LSB_39": 0x39,
    "LSB_3C": 0x3C,
    "LSB_3B": 0x3B,
    "LSB_3E": 0x3E,

    # 3.5 TMS Commands
    "TMS_4A": 0x4A,
    "TMS_4B": 0x4B,
    "TMS_6A": 0x6A,
    "TMS_6B": 0x6B,
    "TMS_6E": 0x6E,
    "TMS_6F": 0x6F,

    # 3.6 Set / Read Data Bits High / Low Bytes
    "SETRDBHLB_80": 0x80,
    "SETRDBHLB_82": 0x82,
    "SETRDBHLB_81": 0x81,
    "SETRDBHLB_83": 0x83,

    # 3.7 Loopback Commands
    "LOOPBACK_EN": 0x84,
    "LOOPBACK_DIS": 0x85,

    # 3.8 Clock Divisor
    # 3.8.1 Set TCK/SK Divisor (FT2232D)
    # 3.8.2 Set clk divisor (FT232H/FT2232H/FT4232H)
    "SET_TCKSK_DIV": 0x86,

    # 4 Instructions for CPU mode
    # 4.2 CPUMode Read Short Address
    "CPU_90": 0x90,
    # 4.3 CPUMode Read Extended Address
    "CPU_91": 0x91,
    # 4.4 CPUMode Write Short Address
    "CPU_92": 0x92,
    # 4.5 CPUMode Write Extended Address
    "CPU_93": 0x93,

    # 5 Instructions for use in both MPSSE and MCU Host Emulation Modes
    # 5.1 Send Immediate
    "SEND_IMMEDIATE": "0x87",
    # 5.2 Wait On I/O High
    "WAIT_IO_HI": "0x88",
    # 5.3 Wait On I/O Low
    "WAIT_IO_LO": "0x89",

    # 6 FT232H, FT2232H & FT4232H ONLY

    # Disables the clk divide by 5 to allow for a 60MHz master clock.
    "TCK_X5": 0x8A,
    # Enables the clk divide by 5 to allow for backward compatibility with FT2232D
    "TCK_D5": 0x8B,
    # Enables 3 phase data clocking. Used by I 2 C interfaces to allow data on both clock edges.
    "ENABLE_3_PHASE_CLOCK": 0x8C,
    # Disables 3 phase data clocking.
    "DISABLE_3_PHASE_CLOCK": 0x8D,
    "CLOCK_N_CYCLES": 0x8E,
    "CLOCK_N8_CYCLES": 0x8F,
    "PULSE_CLOCK_IO_HIGH": 0x94,
    "PULSE_CLOCK_IO_LOW": 0x95,
    "ENABLE_ADAPTIVE_CLOCK": 0x96,
    "DISABLE_ADAPTIVE_CLOCK": 0x97,
    "CLOCK_N8_CYCLES_IO_HIGH": 0x9C,
    "CLOCK_N8_CYCLES_IO_LOW": 0x9D,

    # 7 FT232H ONLY
    # 7.1 Set I/O to only drive on a '0' and tristate on a '1'
    "TRISTATE_IO": 0x9E,

    # basically a nop for testing interface
    "INVALID_COMMAND": 0xAB,
}
mpsse_cmd_i2s = dict([(v, k) for k, v in mpsse_cmd_s2i.items()])

BAD_COMMAND = 0xFA


class MPSSEParser(object):
    def __init__(self):
        self.jo = []

    def next_json(self, j, prefix=None):
        self.jo.append(j)

    def handleRead(self, d):
        print(d['data'])
        buff = bytearray(binascii.unhexlify(d['data']))
        if buff[0] == BAD_COMMAND:
            which = mpsse_cmd_i2s.get(buff[1], None)
            print("read invalid: %s" % which)
        else:
            print("read 0x%02X" % (buff[0], ))

    def handleWrite(self, d):
        print(d['data'])
        buff = bytearray(binascii.unhexlify(d['data']))
        cmd = mpsse_cmd_i2s.get(buff[0], None)
        print(cmd)

    def run(self, j):
        for di, d in enumerate(j["data"]):
            print("")
            print(d)
            if d['type'] == 'read':
                self.handleRead(d)
            elif d['type'] == 'write':
                self.handleWrite(d)
            else:
                print('fixme: %s' % d['type'])

        jret = {
            "data": self.jo,
        }
        return jret


class MPSSETextPrinter(object):
    def __init__(self, args):
        self.ascii = args.ascii

    def next_json(self, j, prefix=None):
        print(j)

    def run(self, j):
        #self.header()

        for d in j["data"]:
            self.next_json(d)

        #self.footer()


class MPSSEJSONSPrinter(object):
    def __init__(self, args):
        self.ascii = args.ascii

    def run(self, j):
        print(json.dumps(j, sort_keys=True, indent=4, separators=(',', ': ')))



from __future__ import print_function
import json
import sys
from . import parsers

indent = ""

print_file = sys.stdout


def default_print_file(fname, cap_file):
    global print_file
    """
    Given argument for explicit file name and a cap file,
    default to the file name, otherwise munge cap file into .py version
    """
    if not fname:
        fname = cap_file.replace('.pcapng',
                                 '.py').replace('.pcap',
                                                '.py').replace('.cap', '.py')

    assert fname != cap_file, (fname, cap_file)
    print_file = open(fname, "w")


def indent_inc():
    global indent

    indent += "    "


def indent_dec():
    global indent

    indent = indent[4:]


def indented(s):
    print("%s%s" % (indent, s), file=print_file)


class Printer(object):
    def __init__(self, argsj):
        self.argsj = argsj

    def run(self, j):
        raise Exception("Required")


class JSONPrinter(Printer):
    def __init__(self, argsj):
        Printer.__init__(self, argsj)

    def run(self, jgen):
        j = parsers.jgen2j(jgen)

        json.dump(j,
                  print_file,
                  sort_keys=True,
                  indent=4,
                  separators=(',', ': '))



from .usb import req2s

setup_reqs = [
    # reply type
    "GET_STATUS",
    "CLEAR_FEATURE",
    "SET_FEATURE",
    "SET_ADDRESS",
    "GET_DESCRIPTOR",
    "SET_DESCRIPTOR",
    "GET_CONFIGURATION",
    "SET_CONFIGURATION",
    "SET_INTERFACE",
    "GET_INTERFACE",
    "CLEAR_FEATURE",
    "SYNCH_FRAME",
]


class SetupFilter(object):
    def __init__(self, argsj, verbose=False):
        # self.setup = argsj.get("setup", False)
        self.verbose = verbose
        self.entries = 0
        self.drops = 0

    def should_filter(self, data):
        return req2s(data["bRequestType"], data["bRequest"]) in setup_reqs

    def gen_data(self, datas):
        for data in datas:
            self.entries += 1
            if data["type"] in ("controlWrite",
                                "controlRead") and self.should_filter(data):
                if self.verbose:
                    print("SetupFilter drop %s (%s %s %s)" %
                          (data['type'],
                           req2s(data["bRequestType"], data["bRequest"]),
                           data["bRequestType"], data["bRequest"]))
                self.drops += 1
                continue
            yield data
        yield {
            "type":
            "comment",
            "v":
            "SetupFilter: dropped %s / %s entries" % (self.drops, self.entries)
        }

    def run(self, jgen):
        for k, v in jgen:
            if k == "data":
                yield k, self.gen_data(v)
            else:
                yield k, v



from .commenter import Commenter
from .setup_filter import SetupFilter
from .vidpid_filter import VidpidFilter
# from .device_filter import DeviceFilter
from .parsers import jgen2j

filters = {
    "commenter": Commenter,
    "setup": SetupFilter,
    "vidpid": VidpidFilter,
    # "device": DeviceFilter,
}


def run(filts, jgen, argsj={}, verbose=False):
    """
    Run series of given filters on input jgen
    First filter in array is applied first
    """
    if len(filts) == 0:
        return jgen

    filt0 = filts[0]
    cls = filters[filt0]
    return run(filts[1:], cls(argsj).run(jgen), argsj=argsj, verbose=verbose)


# FIXME: loaded data is not generated
# This causes issues with filters that expect generated, not JSON data
# ex: "for (k, v) in gen" vs "for (k, v) in gen.items()" conflicts
def run_filter(j, cls, argsj={}, verbose=False):
    c = cls(argsj=argsj, verbose=verbose)
    ret = c.run(j)
    return jgen2j(ret)



from .usb import req2s
import binascii
import struct
from .util import hexdump


def default_arg(argsj, k, default):
    val = argsj.get(k)
    if val is None:
        return default
    else:
        return val


class VidpidFilter(object):
    def __init__(self, argsj, verbose=False):
        # self.setup = argsj.get("setup", False)
        self.verbose = verbose
        self.entries = 0
        self.drops = 0

        self.arg_vid = default_arg(argsj, "vid", None)
        self.arg_pid = default_arg(argsj, "pid", None)
        self.device2vidpid = {}
        self.keep_device = None

    def should_filter(self, data):
        comments = []
        if self.arg_vid is None and self.arg_pid is None:
            return False, comments

        device = data.get('device')
        # Comment / metadata
        if device is None:
            return False, comments

        # a new vid/pid mapping?
        # if data["type"] == "controlRead" and req2s(data["bRequestType"], data["bRequest"]) == "GET_DESCRIPTOR" and data["bDescriptorType"] == 0x01:
        # FIXME: hack
        buff = binascii.unhexlify(data.get("data", ""))
        if data["type"] == "controlRead" and req2s(
                data["bRequestType"],
                data["bRequest"]) == "GET_DESCRIPTOR" and len(buff) == 0x12:
            # not actually decoded
            # TODO: parse this more genericly
            vid, pid = struct.unpack("<HH", buff[0x08:0x0C])
            # print("VID PID 0x%04X 0x%04X" % (vid, pid))
            self.device2vidpid[device] = (vid, pid)
            if (vid == self.arg_vid
                    or self.arg_vid is None) or (pid == self.arg_pid
                                                 or self.arg_pid is None):
                # Note: may appear multiple times
                # First during device 0 enumeration, then once assigned on bus
                # Keep the second one
                if device:
                    if self.keep_device is None:
                        comments.append(
                            self.comment(
                                "VidpidFilter: match device %u w/ 0x%04X:0x%04X"
                                % (device, vid, pid)))
                    elif self.keep_device != device:
                        comments.append(
                            self.comment(
                                "WARNING VidpidFilter: already had different device"
                            ))
                    self.keep_device = device
            return False, comments

        # Filter:
        # Devices not matching target
        # Anything before we've established mapping. Most of this traffic isn't important and simplifies parser
        # print(self.keep_device, device)
        return self.keep_device is None or device != self.keep_device, comments

    def comment(self, s):
        return {
            "type": "comment",
            "v": s,
        }

    def gen_data(self, datas):
        for data in datas:
            self.entries += 1
            should_filter, yields = self.should_filter(data)
            for y in yields:
                yield y
            if should_filter:
                if self.verbose:
                    print("VidpidFilter drop %s (%s %s %s)" %
                          (data['type'],
                           req2s(data["bRequestType"], data["bRequest"]),
                           data["bRequestType"], data["bRequest"]))
                self.drops += 1
                continue
            yield data
        yield self.comment("VidpidFilter: dropped %s / %s entries" %
                           (self.drops, self.entries))

    def run(self, jgen):
        for k, v in jgen:
            if k == "data":
                yield k, self.gen_data(v)
            else:
                yield k, v



import sys
from .pcap_util import PcapParser, load_pcap


def default_arg(argsj, k, default):
    val = argsj.get(k)
    if val is None:
        return default
    else:
        return val


class PcapGen(object):
    def __init__(self, argsj):
        self.verbose = default_arg(argsj, "verbose", False)
        # JSON data objects buffered for next yield
        self.jbuff = None
        self.pcomments = None
        self.min_packet = default_arg(argsj, "min_packet", 0)
        self.max_packet = default_arg(argsj, "max_packet", float('inf'))
        self.cur_packn = None

        # XXX: don't think this is actually used, verify
        self.arg_fx2 = default_arg(argsj, "fx2", False)
        self.arg_device = default_arg(argsj, "device", None)
        # Default to device-hi if no other filtering is specified
        self.arg_device_hi = default_arg(
            argsj, "device_hi", self.arg_device is None
            and not argsj.get("vid") and not argsj.get("pid"))
        self.dev_drops = 0
        self.arg_setup = default_arg(argsj, "setup", False)
        self.arg_halt = default_arg(argsj, "halt", True)
        self.arg_remoteio = default_arg(argsj, "remoteio", False)
        self.arg_rel_pkt = default_arg(argsj, "rel_pkt", False)
        self.arg_print_short = default_arg(argsj, "print_short", False)
        self.arg_comment = default_arg(argsj, "comment", False)
        self.arg_packet_numbers = default_arg(argsj, "packet_numbers", True)

        # Either auto selected or user selected by now
        self.use_pcapng = "pcapng" in argsj["parser"]

    def gcomment(self, s):
        """Add global comment, not attached to a packet"""
        self.jbuff.append({'type': 'comment', 'v': s})

    def gwarning(self, s):
        self.gcomment("WARNING: " + str(s))

    def pcomment(self, s):
        """Add packet comment. Will be attached to the next packet"""
        self.pcomments.append(s)

    def pwarning(self, s):
        self.pcomment("WARNING: " + str(s))

    def printv(self, s):
        if self.verbose:
            print(s)

    def packnum(self):
        '''
        Originally I didn't print anything but found that it was better to keep the line numbers the same
        so that I could diff and then easier back annotate with packet numbers
        '''
        if self.arg_packet_numbers:
            postfix = "%s/%s" % (self.submit.packet_number, self.pktn_str())
        else:
            postfix = "N/A"
        self.pcomment("Generated from packet %s" % postfix)

    def packnumt(self):
        if self.arg_packet_numbers:
            return (self.submit.packet_number, self.pktn_str())
        else:
            return (None, None)

    def gen_data(self):
        parser = PcapParser(self.arg_fin, use_pcapng=self.use_pcapng)
        npackets = 0
        while True:
            npackets += 1
            if not parser.next(self.loop_cb):
                break

            if self.verbose:
                print("gen_data: %s new JSON entries" % (len(self.jbuff), ))

            # Pop packets
            for data in self.jbuff:
                if 0:
                    import json
                    print(data)
                    json.dumps(data)
                yield data
            self.jbuff = []

        if len(self.pending_complete) != 0:
            self.gwarning("%lu pending complete requests" %
                          (len(self.pending_complete)))
        # if len(self.pending_submit) != 0:
        #    self.gwarning("%lu pending submit requests" % (len(self.pending_submit)))

        self.gcomment("PcapGen: generated %u packets" % npackets)
        self.gcomment("PcapGen device filter: dropped %u / %u packets" %
                      (self.dev_drops, npackets))

        # Pop epilogue packets
        for p in self.jbuff:
            yield p
        self.jbuff = []

    def platform(self):
        assert 0, "required"

    def run(self):
        yield 'parser', self.parser()
        yield 'platform', self.platform()
        yield "fn", self.arg_fin
        yield 'args', sys.argv
        yield 'packet_min', self.min_packet
        yield 'packet_max', self.max_packet

        self.jbuff = []
        self.pcomments = []
        self.gcomment("Generated by usbrply")
        self.comment_source()
        self.gcomment("cmd: %s" % (' '.join(sys.argv), ))

        if self.arg_device_hi:
            self.arg_device = -1
            load_pcap(self.arg_fin,
                      self.loop_cb_devmax,
                      use_pcapng=self.use_pcapng)
            self.gcomment('PCapGen device hi: selected device %u' %
                          self.arg_device)
            self.cur_packn = 0

        self.printv("parsing from range %s to %s" %
                    (self.min_packet, self.max_packet))
        yield "data", self.gen_data()



import sys
import subprocess
import json


def clear_screen():
    # 00000000  1b 5b 33 3b 4a 1b 5b 48  1b 5b 32 4a              |.[3;J.[H.[2J|
    sys.stdout.write("\x1b\x5b\x33\x3b\x4a\x1b\x5b\x48\x1b\x5b\x32\x4a")
    sys.stdout.flush()


def hexdump(data, label=None, indent='', address_width=8, f=sys.stdout):
    def isprint(c):
        return c >= ' ' and c <= '~'

    if label:
        print(label)

    bytes_per_half_row = 8
    bytes_per_row = 16
    data = bytearray(data)
    data_len = len(data)

    def hexdump_half_row(start):
        left = max(data_len - start, 0)

        real_data = min(bytes_per_half_row, left)

        f.write(''.join('%02X ' % c for c in data[start:start + real_data]))
        f.write(''.join('   ' * (bytes_per_half_row - real_data)))
        f.write(' ')

        return start + bytes_per_half_row

    pos = 0
    while pos < data_len:
        row_start = pos
        f.write(indent)
        if address_width:
            f.write(('%%0%dX  ' % address_width) % pos)
        pos = hexdump_half_row(pos)
        pos = hexdump_half_row(pos)
        f.write("|")
        # Char view
        left = data_len - row_start
        real_data = min(bytes_per_row, left)

        f.write(''.join([
            c if isprint(c) else '.'
            for c in tostr(data[row_start:row_start + real_data])
        ]))
        f.write((" " * (bytes_per_row - real_data)) + "|\n")


def add_bool_arg(parser, yes_arg, default=False, **kwargs):
    dashed = yes_arg.replace('--', '')
    dest = dashed.replace('-', '_')
    parser.add_argument(yes_arg,
                        dest=dest,
                        action='store_true',
                        default=default,
                        **kwargs)
    kwargs['help'] = 'Disable above'
    parser.add_argument('--no-' + dashed,
                        dest=dest,
                        action='store_false',
                        **kwargs)


# In Python2 bytes_data is a string, in Python3 it's bytes.
# The element type is different (string vs int) and we have to deal
# with that when printing this number as hex.
if sys.version_info[0] == 2:
    myord = ord
else:
    myord = lambda x: x


def tobytes(buff):
    if type(buff) is str:
        #return bytearray(buff, 'ascii')
        return bytearray([myord(c) for c in buff])
    elif type(buff) is bytearray or type(buff) is bytes:
        return buff
    else:
        assert 0, type(buff)


def tostr(buff):
    if type(buff) is str:
        return buff
    elif type(buff) is bytearray or type(buff) is bytes:
        return ''.join([chr(b) for b in buff])
    else:
        assert 0, type(buff)


# Used by scraper scripts
# Due to python2 vs python3 issue, its better to subprocess this
def load_pcap_json(fin, usbrply_args=""):
    if fin.find('.cap') >= 0 or fin.find('.pcapng') >= 0:
        json_fn = '/tmp/scrape.json'
        cmd = 'usbrply %s  --json %s >%s' % (usbrply_args, fin, json_fn)
        subprocess.check_call(cmd, shell=True)
    else:
        json_fn = fin

    j = json.load(open(json_fn))
    return j, json_fn



from __future__ import print_function
from usbrply import printer
from .printer import Printer, indented, indent_inc, indent_dec
import sys
import binascii
from . import usb
from .util import myord


def comment(s):
    indented('# %s' % (s, ))


def bytes2AnonArray(bytes_data):
    # In Python2 bytes_data is a string, in Python3 it's bytes.
    # The element type is different (string vs int) and we have to deal
    # with that when printing this number as hex.

    byte_str = "b\""

    for i in range(len(bytes_data)):
        if i and i % 16 == 0:
            byte_str += '\"\n            b\"'
        byte_str += "\\x%02X" % (myord(bytes_data[i]), )
    return byte_str + "\""


def printControlRequest(submit, data_str, data_size, pipe_str):
    if submit.m_ctrl.bRequestType & usb.URB_TRANSFER_IN:
        indented("buff = controlRead(0x%02X, 0x%02X, 0x%04X, 0x%04X, %u)" %
                 (submit.m_ctrl.bRequestType, submit.m_ctrl.bRequest,
                  submit.m_ctrl.wValue, submit.m_ctrl.wIndex, data_size))
    else:
        indented("controlWrite(0x%02X, 0x%02X, 0x%04X, 0x%04X, %s)" %
                 (submit.m_ctrl.bRequestType, submit.m_ctrl.bRequest,
                  submit.m_ctrl.wValue, submit.m_ctrl.wIndex, data_str))


class LibusbPyPrinter(Printer):
    def __init__(self, argsj, verbose=False):
        Printer.__init__(self, argsj)
        self.prevd = None
        self.wrapper = argsj.get("wrapper", False)
        self.sleep = argsj.get("sleep", False)
        self.packet_numbers = argsj.get("packet_numbers", True)
        # FIXME
        self.vid = 0
        self.pid = 0
        self.verbose = verbose

    def print_imports(self):
        print('''\
import binascii
import time
import usb1
''',
              file=printer.print_file)

    def print_wrapper_header(self):
        print('''\
def validate_read(expected, actual, msg):
    if expected != actual:
        print('Failed %s' % msg)
        print('  Expected; %s' % binascii.hexlify(expected,))
        print('  Actual:   %s' % binascii.hexlify(actual,))
        #raise Exception('failed validate: %s' % msg)

''',
              file=printer.print_file)
        print('def replay(dev):', file=printer.print_file)
        indent_inc()
        print('''\
    def bulkRead(endpoint, length, timeout=None):
        return dev.bulkRead(endpoint, length, timeout=(1000 if timeout is None else timeout))

    def bulkWrite(endpoint, data, timeout=None):
        dev.bulkWrite(endpoint, data, timeout=(1000 if timeout is None else timeout))
    
    def controlRead(bRequestType, bRequest, wValue, wIndex, wLength,
                    timeout=None):
        return dev.controlRead(bRequestType, bRequest, wValue, wIndex, wLength,
                    timeout=(1000 if timeout is None else timeout))

    def controlWrite(bRequestType, bRequest, wValue, wIndex, data,
                     timeout=None):
        dev.controlWrite(bRequestType, bRequest, wValue, wIndex, data,
                     timeout=(1000 if timeout is None else timeout))

    def interruptRead(endpoint, size, timeout=None):
        return dev.interruptRead(endpoint, size,
                    timeout=(1000 if timeout is None else timeout))

    def interruptWrite(endpoint, data, timeout=None):
        dev.interruptWrite(endpoint, data, timeout=(1000 if timeout is None else timeout))
''',
              file=printer.print_file)

    def header(self):
        comment("Generated by usbrply")
        # comment("Date: %s" % (UVDCurDateTime()))
        comment("cmd: %s" % (' '.join(sys.argv), ))
        indented("")

        if self.wrapper:
            self.print_imports()
            print("", file=printer.print_file)
            self.print_wrapper_header()

    def footer(self):
        if not self.wrapper:
            return
        print('''
def open_dev(usbcontext=None):
    if usbcontext is None:
        usbcontext = usb1.USBContext()
    
    print('Scanning for devices...')
    for udev in usbcontext.getDeviceList(skip_on_error=True):
        vid = udev.getVendorID()
        pid = udev.getProductID()
        if (vid, pid) == (''' + "0x%04X, 0x%04X" % (self.vid, self.pid) + '''):
            print("")
            print("")
            print('Found device')
            print('Bus %03i Device %03i: ID %04x:%04x' % (
                udev.getBusNumber(),
                udev.getDeviceAddress(),
                vid,
                pid))
            return udev.open()
    raise Exception("Failed to find a device")

if __name__ == "__main__":
    import argparse 
    
    parser = argparse.ArgumentParser(description='Replay captured USB packets')
    args = parser.parse_args()

    usbcontext = usb1.USBContext()
    dev = open_dev(usbcontext)
    dev.claimInterface(0)
    dev.resetDevice()
    replay(dev)

''',
              file=printer.print_file)

    def packet_number_str(self, d):
        if self.packet_numbers:
            return "packet %s/%s" % (d["submit"]["packn"],
                                     d["complete"]["packn"])
        else:
            # TODO: consider counting instead of by captured index
            return "packet"

    def parse_data(self, d):
        # print(d)
        if self.sleep and self.prevd and d["type"] != "comment":
            try:
                dt = d["submit"]["t"] - self.prevd["submit"]["t"]
            except KeyError:
                raise ValueError("Input JSON does not support timestamps")
            if dt >= 0.001:
                indented('time.sleep(%.3f)' % (dt, ))

        if d["type"] == "comment":
            comment(d["v"])
            return

        packet_numbering = self.packet_number_str(d)

        if "comments" in d:
            for c in d["comments"]:
                comment(c)

        if d["type"] == "controlRead":
            # Is it legal to have a 0 length control in?
            indented("buff = controlRead(0x%02X, 0x%02X, 0x%04X, 0x%04X, %u)" %
                     (d["bRequestType"], d["bRequest"], d["wValue"],
                      d["wIndex"], d["wLength"]))
            indented("validate_read(%s, buff, \"%s\")" % (bytes2AnonArray(
                binascii.unhexlify(d["data"])), packet_numbering))
        elif d["type"] == "controlWrite":
            data_str = bytes2AnonArray(binascii.unhexlify(d["data"]))
            indented("controlWrite(0x%02X, 0x%02X, 0x%04X, 0x%04X, %s)" %
                     (d["bRequestType"], d["bRequest"], d["wValue"],
                      d["wIndex"], data_str))

        elif d["type"] == "bulkRead":
            data_str = "\"\""
            indented("buff = bulkRead(0x%02X, 0x%04X)" % (d["endp"], d["len"]))
            indented("validate_read(%s, buff, \"%s\")" % (bytes2AnonArray(
                binascii.unhexlify(d["data"])), packet_numbering))
        elif d["type"] == "bulkWrite":
            # Note that its the submit from earlier, not the ack that we care about
            data_str = bytes2AnonArray(binascii.unhexlify(d["data"]))
            # def bulkWrite(self, endpoint, data, timeout=0):
            indented("bulkWrite(0x%02X, %s)" % (d["endp"], data_str))

        elif d["type"] == "interruptRead":
            data_str = "\"\""
            indented("buff = interruptRead(0x%02X, 0x%04X)" %
                     (d["endp"], d["len"]))
            indented("validate_read(%s, buff, \"%s\")" % (bytes2AnonArray(
                binascii.unhexlify(d["data"])), packet_numbering))

        elif d["type"] == "interruptWrite":
            indented("interruptWrite(0x%02X, %s)" % (d["endp"], data_str))
        else:
            if self.verbose:
                print("LibusbPyPrinter WARNING: dropping %s" % (d["type"], ))

        # these aren't event added to JSON right now
        # print('%s# WARNING: omitting interrupt' % (indent,))

        if d["type"] != "comment":
            self.prevd = d

    def run(self, jgen):
        self.header()

        # Last wire command (ie non-comment)
        # Used to optionally generate timing
        self.prevd = None

        # Convert generator into static JSON
        j = {}
        for k, v in jgen:
            j[k] = v

        for d in j["data"]:
            self.parse_data(d)

        self.footer()



from .usb import req2s, feat_i2s
import binascii


class Commenter(object):
    def __init__(self, argsj, verbose=False):
        self.fx2 = argsj.get("fx2", False)
        self.verbose = verbose

    # FX2 regs: http://www.keil.com/dd/docs/datashts/cypress/fx2_trm.pdf
    # FX2LP regs: http://www.cypress.com/file/126446/download
    def try_comment(self, data):
        # Table 9-3. Standard Device Requests
        vendor = None
        if self.fx2:
            vendor = {
                0xA0: "FX2_REG_W",
            }
        reqs = req2s(data["bRequestType"], data["bRequest"], vendor)
        if not reqs:
            return

        ret = '%s (0x%02X)' % (reqs, data["bRequest"])
        if reqs == 'SET_ADDRESS':
            ret += ': 0x%02x/%d' % (data["wValue"], data["wValue"])
        elif reqs == 'SET_FEATURE' or reqs == 'CLEAR_FEATURE':
            ret += ': 0x%02X (%s)' % (data["wValue"],
                                      feat_i2s.get(data["wValue"], 'unknown'))
        elif reqs == 'FX2_REG_W':
            addr = data["wValue"]
            reg2s = {
                0xE600: 'CPUCS',
            }
            reg = reg2s.get(addr, None)

            # 5.4 FX2 Memory Maps
            # Appendix C
            # FX2 Register Summary
            ret += ': addr=0x%04X' % (data["wValue"])
            if reg:
                ret += ' (%s)' % (reg, )
            elif addr < 0x1000:
                ret += ' (FW load)'
            # FX2: 8K of on-chip RAM (the "Main RAM") at addresses 0x0000-0x1FFF
            # FX2LP: 16K
            elif 0x0000 <= addr <= 0x3FFF:
                ret += ' (main RAM addr=0x%04X)' % addr
            # 512 bytes of on-chip RAM (the "Scratch RAM") at addresses 0xE000-0xE1FFF
            elif 0xE000 <= addr <= 0xE1FF:
                ret += ' (scratch RAM)'
            # The CPU communicates with the SIE using a set of registers occupying on-chip RAM addresses 0xE600-0xE6FF"
            elif 0xE600 <= addr <= 0xE6FF:
                ret += ' (unknown reg)'
            # per memory map: 7.5KB of USB regs and 4K EP buffers
            elif 0xE200 <= addr <= 0xFFFF:
                ret += ' (unknown misc)'
            else:
                ret += ' (unknown)'

            bdat = binascii.unhexlify(data["data"])
            if len(bdat) == 1:
                dat = ord(bdat)
                if reg == 'CPUCS':
                    if dat & 1 == 1:
                        ret += ', reset: hold'
                    else:
                        ret += ', reset: release'
        l = data.get("comments", [])
        l.append(ret)
        data["comments"] = l

    def gen_data(self, datas):
        for data in datas:
            if data["type"] in ("controlWrite", "controlRead"):
                self.try_comment(data)
            yield data

    def run(self, jgen):
        for k, v in jgen:
            if k == "data":
                yield k, self.gen_data(v)
            else:
                yield k, v



#!/usr/bin/env python

import usbrply.parsers
import usbrply.printers
import unittest
import os
from usbrply import printer
from usbrply import parsers
import warnings

devnull = open("/dev/null", "w")


class TestCase(unittest.TestCase):
    def setUp(self):
        """Call before every test case."""
        warnings.simplefilter("ignore")
        printer.print_file = devnull

    def tearDown(self):
        """Call after every test case."""
        pass

    def test_parse_win_pcap(self):
        """Windows .pcap parse test"""
        parsers.jgen2j(usbrply.parsers.pcap2json("test/data/win1.pcapng"))

    def test_parse_lin_pcap(self):
        """Linux .pcap parse test"""
        parsers.jgen2j(usbrply.parsers.pcap2json("test/data/lin1.pcapng"))

    def test1(self):
        return
        j = parsers.jgen2j(usbrply.parsers.pcap2json("test/data/win1.pcapng"))
        for d in j["data"]:
            if d["type"] == "controlRead":
                print(d)

    def test_print_json(self):
        usbrply.printers.run(
            "json", usbrply.parsers.pcap2json("test/data/lin1.pcapng"))

    def test_print_pyprinter_lin(self):
        usbrply.printers.run(
            "libusb-py", usbrply.parsers.pcap2json("test/data/lin1.pcapng"))

    def test_print_pyprinter_win(self):
        usbrply.printers.run(
            "libusb-py", usbrply.parsers.pcap2json("test/data/win1.pcapng"))

    def test_win_interrupt(self):
        usbrply.printers.run(
            "json",
            usbrply.parsers.pcap2json("test/data/win_interrupt.pcapng"))


if __name__ == "__main__":
    unittest.main()  # run all tests



import os
from setuptools import setup, find_packages
import shutil


# Utility function to read the README file.
# Used for the long_description.  It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

if not os.path.exists('build'):
    os.mkdir('build')
scripts = (('main.py', 'usbrply'),
           #('serial.py', 'usbrply-serial'),
           )
scripts_dist = []
for src_fn, prog_name in scripts:
    # Make script names more executable like
    dst = 'build/' + prog_name
    #shutil.copy(src_fn, dst)
    if os.path.exists(dst):
        os.unlink(dst)
    os.symlink(os.path.realpath(src_fn), dst)
    scripts_dist.append(dst)

setup(
    name="usbrply",
    version="2.1.1",
    author="John McMaster",
    author_email='JohnDMcMaster@gmail.com',
    description=("Replay captured USB packets from .pcap file."),
    license="BSD",
    keywords="libusb pcap",
    url='https://github.com/JohnDMcMaster/usbrply',
    packages=find_packages(),
    scripts=scripts_dist,
    install_requires=[
        "python-pcapng",
    ],
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    classifiers=[
        "License :: OSI Approved :: BSD License",
    ],
)



#!/usr/bin/env python

from .usb import *
from .util import hexdump, tostr
from .com_pcap import PcapGen

import sys
import binascii
import struct
from collections import namedtuple
import os
import errno


# When we get an IN request we may process packets in between
class PendingRX:
    def __init__(self):
        # Unprocessed packet bytes

        self.raw = None
        #usb_urb_t m_urb
        self.m_urb = None
        #usb_ctrlrequest m_ctrl
        # Only applies to control requests
        self.m_ctrl = None
        self.packet_number = 0

        # uint8_t *m_data_out
        self.m_data_out = None


class payload_bytes_type_t:
    def __init__(self):
        self.req_in = 0
        self.req_in_last = None
        self.in_ = 0
        self.in_last = None

        self.req_out = 0
        self.req_out_last = None
        self.out = 0
        self.out_last = None


def update_delta(pb):
    pb.req_in_last = pb.req_in
    pb.in_last = pb.in_

    pb.req_out_last = pb.req_out
    pb.out_last = pb.out


'''
struct usb_ctrlrequest {
    __u8 bRequestType;
    __u8 bRequest;
    __le16 wValue;
    __le16 wIndex;
    __le16 wLength;
} __attribute__ ((packed));
'''
usb_ctrlrequest_nt = namedtuple(
    'usb_ctrlrequest',
    (
        'bRequestType',
        'bRequest',
        'wValue',
        'wIndex',
        'wLength',
        # FIXME: what exactly are these?
        'res'))
usb_ctrlrequest_fmt = '<BBHHHH'
usb_ctrlrequest_sz = struct.calcsize(usb_ctrlrequest_fmt)


def usb_ctrlrequest(s):
    return usb_ctrlrequest_nt(*struct.unpack(usb_ctrlrequest_fmt, bytes(s)))


def bytes2Hex(bytes_data):
    return tostr(binascii.hexlify(bytes_data))


def deviceStr():
    # return "dev.udev"
    return "udev"


def print_urb(urb):
    print("URB id: 0x%016lX" % (urb.id))
    print("  type: %s (%c / 0x%02X)" %
          (urb_type2str[urb.type], urb.type, urb.type))
    #print("    dir: %s" % ('IN' if urb.type & URB_TRANSFER_IN else 'OUT',))
    print("  transfer_type: %s (0x%02X)" %
          (transfer2str[urb.transfer_type], urb.transfer_type))
    print("  endpoint: 0x%02X" % (urb.endpoint))
    print("  device: 0x%02X" % (urb.device))
    print("  bus_id: 0x%04X" % (urb.bus_id))
    print("  setup_request: 0x%02X" % (urb.setup_request))
    print("  data: 0x%02X" % (urb.data))
    #print("  sec: 0x%016llX" % (urb.sec))
    print("  usec: 0x%08X" % (urb.usec))
    print("  status: 0x%08X" % (urb.status))
    print("  length: 0x%08X" % (urb.length))
    print("  data_length: 0x%08X" % (urb.data_length))


def urb2json(urb):
    j = dict(urb._asdict())
    j["ctrlrequest"] = bytes2Hex(j["ctrlrequest"])
    # j["data"] = binascii.hexlify(j["data"])
    j["t"] = j['sec'] + j['usec'] / 1e6
    return j


'''
typedef struct {
    uint64_t id
    uint8_t type
    uint8_t transfer_type
    uint8_t endpoint
    uint8_t device
    uint16_t bus_id
    uint8_t setup_request
    uint8_t data
    uint64_t sec
    uint32_t usec
    uint32_t status
    uint32_t length
    uint32_t data_length
    //These form the URB setup for control transfers and are techincally part of the URB
    //uint8_t pad[24]
} __attribute__((packed)) usb_urb_t
'''
usb_urb_nt = namedtuple(
    'usb_urb',
    (
        'id',
        'type',
        'transfer_type',
        'endpoint',
        'device',
        'bus_id',
        'setup_request',
        'data',
        'sec',
        'usec',
        'status',

        # Control: requested data length
        # Complete: identical to data_length?
        'length',
        # How much data is attached to this message
        'data_length',
        # Main use is URB setup for control requests, not sure if thats universal?
        # TODO: check how these are used in bulk requests
        # If it is the same, they should be merged into this structure
        'ctrlrequest',
    ))
usb_urb_fmt = (
    '<'
    'Q'  # id
    'B'
    'B'
    'B'
    'B'  # device
    'H'
    'B'
    'B'
    'Q'  # sec
    'I'
    'i'
    'I'  # length
    'I'
    '24s')
usb_urb_sz = struct.calcsize(usb_urb_fmt)


def usb_urb(s):
    return usb_urb_nt(*struct.unpack(usb_urb_fmt, bytes(s)))


class Gen(PcapGen):
    def __init__(self, fn, argsj={}):
        PcapGen.__init__(self, argsj)

        self.arg_fin = fn
        self.cur_packn = 0
        self.rel_pkt = 0

        # Pending requests
        # Typically size 0-1 but sometimes more pile up
        # Entry can be set to None which means drop on complete
        # Holds PendingRXs
        self.pending_complete = {}

    """
    @property
    def arg_setup(self):
        return bool(self.args.get("setup", True))
    """

    def comment_source(self):
        self.gcomment('Source: Linux pcap (usbmon)')

    def parser(self):
        return "lin-pcap"

    def platform(self):
        return "linux"

    def loop_cb_devmax(self, caplen, packet, ts):
        self.cur_packn += 1
        if self.cur_packn < self.min_packet or self.cur_packn > self.max_packet:
            # print("# Skipping packet %d" % (self.cur_packn))
            return

        if caplen != len(packet):
            print("packet %s: malformed, caplen %d != len %d", self.pktn_str(),
                  caplen, len(packet))
            return
        if self.verbose:
            print('Len: %d' % len(packet))

        self.printv("Length %u" % (len(packet), ))
        if len(packet) < usb_urb_sz:
            self.verbose and hexdump(packet)
            raise ValueError("Packet size %d is not min size %d" %
                             (len(packet), usb_urb_sz))

        # caplen is actual length, len is reported
        self.urb_raw = packet
        self.urb = usb_urb(packet[0:usb_urb_sz])
        dat_cur = packet[usb_urb_sz:]

        self.arg_device = max(self.arg_device, self.urb.device)

    def loop_cb(self, caplen, packet, ts):
        packet = bytearray(packet)
        self.cur_packn += 1
        if self.cur_packn < self.min_packet or self.cur_packn > self.max_packet:
            # print("# Skipping packet %d" % (self.cur_packn))
            return
        if self.verbose:
            print("")
            print("")
            print("")
            print('LIN PACKET %s' % (self.cur_packn, ))

        if caplen != len(packet):
            print("packet %s: malformed, caplen %d != len %d", self.pktn_str(),
                  caplen, len(packet))
            return
        if self.verbose:
            print('Len: %d' % len(packet))

        self.printv("Length %u" % (len(packet), ))
        if len(packet) < usb_urb_sz:
            hexdump(packet)
            raise ValueError("Packet size %d is not min size %d" %
                             (len(packet), usb_urb_sz))

        # caplen is actual length, len is reported
        self.urb_raw = packet
        self.urb = usb_urb(packet[0:usb_urb_sz])
        dat_cur = packet[usb_urb_sz:]

        # Main packet filtering
        # Drop if not specified device
        if self.arg_device is not None and self.urb.device != self.arg_device:
            if self.verbose:
                print('packet %s: want device %s, got %s' %
                      (self.pktn_str(), self.arg_device, self.urb.device))
            self.dev_drops += 1
            return
        self.rel_pkt += 1

        if self.verbose:
            print("Header size: %lu" % (usb_urb_sz, ))
            print_urb(self.urb)

        if self.urb.type == URB_ERROR:
            print("oh noes!")
            if self.arg_halt:
                sys.exit(1)

        if self.urb.type == URB_COMPLETE:
            if self.verbose:
                print('Pending completes (%d):' %
                      (len(self.pending_complete), ))
                for k in self.pending_complete:
                    print('  0x%016lX' % (k, ))
            # for some reason usbmon will occasionally give packets out of order
            if not self.urb.id in self.pending_complete:
                # Interrupts can generate these
                if self.verbose:
                    self.gwarning(
                        "Packet %s missing submit.  URB ID: 0x%016lX" %
                        (self.pktn_str(), self.urb.id))
            else:
                self.process_complete(self.pending_complete[self.urb.id],
                                      self.urb, dat_cur)

        elif self.urb.type == URB_SUBMIT:
            # Find the matching submit request
            if self.urb.transfer_type == URB_CONTROL:
                self.processControlSubmit(dat_cur)
            elif self.urb.transfer_type == URB_BULK:
                self.processBulkSubmit(dat_cur)
            elif self.urb.transfer_type == URB_INTERRUPT:
                pending = PendingRX()
                pending.raw = self.urb_raw
                pending.m_urb = self.urb
                pending.packet_number = self.pktn_str()
                if self.verbose:
                    print('Added pending interrupt URB 0x%016lX' % self.urb.id)
                self.pending_complete[self.urb.id] = pending

        # Should have either generated no comments or attached them
        assert len(self.pcomments) == 0, ("unhandled comment", self.pcomments)
        self.submit = None
        self.urb = None

    def pktn_str(self):
        if self.arg_rel_pkt:
            return self.rel_pkt
        else:
            return self.cur_packn

    def process_complete(self, pending_rx, urb_complete, dat_cur):
        """
        Warning: this may be called with current urb as either the submit or the complete
        Normalize here which may swap self.urb
        """
        # assert type(pending_rx) is PendingRX, type(pending_rx)
        self.submit = pending_rx
        self.urb = urb_complete

        # Discarded?
        if self.submit is not None:
            self.packnum()

            # What was EREMOTEIO?
            EREMOTEIO = -121
            if self.urb.status != 0 and not (not self.arg_remoteio
                                             and self.urb.status == EREMOTEIO):
                # Interrupts can generate these
                if self.verbose:
                    self.gwarning(
                        'Packet %s complete code %s (%s)' %
                        (self.submit.packet_number, self.urb.status,
                         errno.errorcode.get(-self.urb.status, "unknown")))

            # Find the matching submit request
            if self.urb.transfer_type == URB_CONTROL:
                if self.submit.m_ctrl.bRequestType & URB_TRANSFER_IN:
                    self.processControlCompleteIn(dat_cur)
                else:
                    self.processControlCompleteOut(dat_cur)
            elif self.urb.transfer_type == URB_BULK:
                if self.urb.endpoint & USB_DIR_IN:
                    self.processBulkCompleteIn(dat_cur)
                else:
                    self.processBulkCompleteOut(dat_cur)
            elif self.urb.transfer_type == URB_INTERRUPT:
                if self.urb.endpoint & URB_TRANSFER_IN:
                    self.processInterruptCompleteIn(dat_cur)
                else:
                    self.processInterruptCompleteOut(dat_cur)
            else:
                if self.verbose:
                    print("WARNING: unhandled transfer type %s" %
                          (self.urb.transfer_type == URB_INTERRUPT, ))

        if self.urb.id in self.pending_complete:
            del self.pending_complete[self.urb.id]

    def processControlSubmit(self, dat_cur):
        pending = PendingRX()
        pending.raw = self.urb_raw
        pending.m_urb = self.urb

        if self.verbose:
            print('Remaining data: %d' % (len(dat_cur)))
            print('ctrlrequest: %d' % (len(self.urb.ctrlrequest)))
        ctrl = usb_ctrlrequest(self.urb.ctrlrequest[0:usb_ctrlrequest_sz])

        if self.verbose:
            print("Packet %s control submit (control info size %lu)" %
                  (self.pktn_str(), 666))
            print("    bRequestType: %s (0x%02X)" %
                  (request_type2str(ctrl.bRequestType), ctrl.bRequestType))
            #print("    bRequest: %s (0x%02X)" % (request2str(ctrl), ctrl.bRequest))
            print("    wValue: 0x%04X" % (ctrl.wValue))
            print("    wIndex: 0x%04X" % (ctrl.wIndex))
            print("    wLength: 0x%04X" % (ctrl.wLength))

        if (ctrl.bRequestType & URB_TRANSFER_IN) == URB_TRANSFER_IN:
            self.printv("%d: IN" % (self.cur_packn))
        else:
            self.printv("%d: OUT" % (self.cur_packn))
            if len(dat_cur) != self.urb.data_length:
                self.gwarning(
                    "remaining bytes %d != expected payload out bytes %d" %
                    (len(dat_cur), self.urb.data_length))
                hexdump(dat_cur, "  ")
                #raise Exception('See above')
            pending.m_data_out = bytes(dat_cur)

        pending.m_ctrl = ctrl
        pending.packet_number = self.pktn_str()
        if self.verbose:
            print('Added pending control URB %s' % self.urb.id)
        self.pending_complete[self.urb.id] = pending

    def processControlCompleteIn(self, dat_cur):
        packet_numbering = ''
        data_size = 0
        data_str = "None"
        max_payload_sz = self.submit.m_ctrl.wLength

        # Is it legal to have a 0 length control in?
        if self.submit.m_ctrl.wLength:
            data_str = "buff"
            data_size = self.submit.m_ctrl.wLength

        # Verify we actually have enough / expected
        # If exact match don't care
        if len(dat_cur) != max_payload_sz:
            if len(dat_cur) < max_payload_sz:
                if self.arg_print_short:
                    self.pcomment("NOTE:: req max %u but got %u" %
                                  (max_payload_sz, len(dat_cur)))
            else:
                raise Exception('invalid response')

        self.output_packet({
            'type': 'controlRead',
            'bRequestType': self.submit.m_ctrl.bRequestType,
            'bRequest': self.submit.m_ctrl.bRequest,
            'wValue': self.submit.m_ctrl.wValue,
            'wIndex': self.submit.m_ctrl.wIndex,
            'wLength': self.submit.m_ctrl.wLength,
            'data': bytes2Hex(dat_cur)
        })

        if self.submit.m_ctrl.wLength:
            if self.arg_packet_numbers:
                packet_numbering = "packet %s/%s" % (self.submit.packet_number,
                                                     self.pktn_str())
            else:
                # TODO: consider counting instead of by captured index
                packet_numbering = "packet"

    def processControlCompleteOut(self, dat_cur):
        data_size = 0
        data_str = "None"

        #print('Control out w/ len %d' % len(submit.m_data_out))

        # print("Data out size: %u vs urb size %u" % (submit.m_data_out_size, submit.m_urb.data_length ))
        if len(self.submit.m_data_out):
            # Note that its the submit from earlier, not the ack that we care about
            data_str = bytes2Hex(self.submit.m_data_out)
            data_size = len(self.submit.m_data_out)

        self.output_packet({
            'type': 'controlWrite',
            'bRequestType': self.submit.m_ctrl.bRequestType,
            'bRequest': self.submit.m_ctrl.bRequest,
            'wValue': self.submit.m_ctrl.wValue,
            'wIndex': self.submit.m_ctrl.wIndex,
            'data': bytes2Hex(self.submit.m_data_out)
        })

    def output_packet(self, j):
        urbj_submit = urb2json(self.submit.m_urb)
        urbj_complete = urb2json(self.urb)
        nsub, ncomplete = self.packnumt()
        j["submit"] = {
            "packn": nsub,
            'urb': urbj_submit,
            't': urbj_submit["t"],
        }
        j["device"] = j["submit"]["urb"]["device"]
        j["complete"] = {
            'packn': ncomplete,
            'urb': urbj_complete,
            't': urbj_complete["t"],
        }
        if len(self.pcomments):
            j["comments"] = self.pcomments
        self.jbuff.append(j)
        self.pcomments = []

    def processBulkSubmit(self, dat_cur):
        pending = PendingRX()
        pending.raw = self.urb_raw
        pending.m_urb = self.urb

        if self.verbose:
            print('Remaining data: %d' % (len(dat_cur)))

        #if self.verbose:
        #    print("Packet %d bulk submit (control info size %lu)" % (self.pktn_str(), 666))

        if self.urb.endpoint & URB_TRANSFER_IN:
            self.printv("%d: IN" % (self.cur_packn))
        else:
            self.printv("%d: OUT" % (self.cur_packn))
            if len(dat_cur) != self.urb.data_length:
                self.pwarning(
                    "remaining bytes %d != expected payload out bytes %d" %
                    (len(dat_cur), self.urb.data_length))
                hexdump(dat_cur, "  ")
                #raise Exception('See above')
            pending.m_data_out = bytes(dat_cur)

        pending.packet_number = self.pktn_str()
        self.pending_complete[self.urb.id] = pending
        if self.verbose:
            print('Added pending bulk URB 0x%016lX' % self.urb.id)

    def processBulkCompleteIn(self, dat_cur):
        # packet_numbering = ''
        data_size = 0
        data_str = "None"
        max_payload_sz = self.submit.m_urb.length

        # FIXME: this is a messy conversion artfact from the C code
        # Is it legal to have a 0 length bulk in?
        if max_payload_sz:
            data_str = "buff"
            data_size = max_payload_sz

        self.output_packet({
            'type': 'bulkRead',
            'endp': self.submit.m_urb.endpoint,
            'len': data_size,
            'data': bytes2Hex(dat_cur)
        })

        # Verify we actually have enough / expected
        # If exact match don't care
        if len(dat_cur) > max_payload_sz:
            self.pwarning('requested max %u bytes but got %u' %
                          (max_payload_sz, len(dat_cur)))
        elif len(dat_cur) < max_payload_sz and self.arg_print_short:
            self.pcomment("NOTE:: req max %u but got %u" %
                          (max_payload_sz, len(dat_cur)))
        """
        if max_payload_sz:
            if args.packet_numbers:
                packet_numbering = "packet %s/%s" % (self.submit.packet_number,
                                                     self.pktn_str())
            else:
                # TODO: consider counting instead of by captured index
                packet_numbering = "packet"
        """

    def processBulkCompleteOut(self, dat_cur):
        # output below
        self.output_packet({
            'type': 'bulkWrite',
            'endp': self.submit.m_urb.endpoint,
            'data': bytes2Hex(self.submit.m_data_out)
        })

    def processInterruptCompleteOut(self, dat_cur):
        # looks like maybe windows doesn't report the request size?
        # think this is always 0
        # assert self.submit.m_urb.length == 0

        # FIXME: this is a messy conversion artifact from the C code
        # Is it legal to have a 0 length bulk in?
        data_size = 0
        # instead, use the recieved buffer size as a best estimate
        max_payload_sz = len(dat_cur)
        if max_payload_sz:
            data_size = max_payload_sz

        self.output_packet({
            'type': 'interruptWrite',
            'endp': self.submit.m_urb.endpoint,
            'len': data_size,
            'data': bytes2Hex(dat_cur)
        })

    def processInterruptCompleteIn(self, dat_cur):
        # looks like maybe windows doesn't report the request size?
        # think this is always 0
        # assert self.submit.m_urb.length == 0

        # FIXME: this is a messy conversion artifact from the C code
        # Is it legal to have a 0 length bulk in?
        data_size = 0
        # instead, use the recieved buffer size as a best estimate
        max_payload_sz = len(dat_cur)
        if max_payload_sz:
            data_size = max_payload_sz

        # output below
        self.output_packet({
            'type': 'interruptRead',
            'endp': self.submit.m_urb.endpoint,
            'len': data_size,
            'data': bytes2Hex(dat_cur)
        })



'''
possible transfer mode
'''
URB_TRANSFER_IN = 0x80
URB_ISOCHRONOUS = 0x0
URB_INTERRUPT = 0x1
URB_CONTROL = 0x2
URB_BULK = 0x3
'''
possible event type
'''
URB_SUBMIT = ord('S')
URB_COMPLETE = ord('C')
URB_ERROR = ord('E')

urb_type2str = {
    URB_SUBMIT: 'URB_SUBMIT',
    URB_COMPLETE: 'URB_COMPLETE',
    URB_ERROR: 'URB_ERROR',
}

USB_REQ_GET_STATUS = 0x00
USB_REQ_CLEAR_FEATURE = 0x01
# 0x02 is reserved
USB_REQ_SET_FEATURE = 0x03
# 0x04 is reserved
USB_REQ_SET_ADDRESS = 0x05
USB_REQ_GET_DESCRIPTOR = 0x06
USB_REQ_SET_DESCRIPTOR = 0x07
USB_REQ_GET_CONFIGURATION = 0x08
USB_REQ_SET_CONFIGURATION = 0x09
USB_REQ_GET_INTERFACE = 0x0A
USB_REQ_SET_INTERFACE = 0x0B
USB_REQ_SYNCH_FRAME = 0x0C

USB_DIR_OUT = 0  # to device
USB_DIR_IN = 0x80  # to host

USB_TYPE_MASK = (0x03 << 5)  # 0x60
USB_TYPE_STANDARD = (0x00 << 5)  # 0x00
USB_TYPE_CLASS = (0x01 << 5)  # 0x20
USB_TYPE_VENDOR = (0x02 << 5)  # 0x40
USB_TYPE_RESERVED = (0x03 << 5)  # 0x60

USB_RECIP_MASK = 0x1f
USB_RECIP_DEVICE = 0x00
USB_RECIP_INTERFACE = 0x01
USB_RECIP_ENDPOINT = 0x02
USB_RECIP_OTHER = 0x03
# From Wireless USB 1.0
USB_RECIP_PORT = 0x04
USB_RECIP_RPIPE = 0x05

# Table 9-6. Standard Feature Selectors
feat_i2s = {
    # Endpoint
    0: 'ENDPOINT_HALT',
    # Device
    1: 'DEVICE_REMOTE_WAKEUP',
    # Device
    2: 'TEST_MODE',
}


def req2s(bRequestType, bRequest, vendor=None):
    m = {
        USB_TYPE_STANDARD: {
            USB_REQ_GET_STATUS: "GET_STATUS",
            USB_REQ_CLEAR_FEATURE: "CLEAR_FEATURE",
            USB_REQ_SET_FEATURE: "SET_FEATURE",
            USB_REQ_SET_ADDRESS: "SET_ADDRESS",
            USB_REQ_GET_DESCRIPTOR: "GET_DESCRIPTOR",
            USB_REQ_SET_DESCRIPTOR: "SET_DESCRIPTOR",
            USB_REQ_GET_CONFIGURATION: "GET_CONFIGURATION",
            USB_REQ_SET_CONFIGURATION: "SET_CONFIGURATION",
            USB_REQ_SET_INTERFACE: "SET_INTERFACE",
        },
        USB_TYPE_CLASS: {
            USB_REQ_GET_STATUS: "GET_STATUS",
            USB_REQ_CLEAR_FEATURE: "CLEAR_FEATURE",
            USB_REQ_SET_FEATURE: "SET_FEATURE",
            USB_REQ_GET_INTERFACE: "GET_INTERFACE",
        },
        # 2020-12-29: these are interfering with a device
        # Leave these out for now, try to figure out why they were added
        # and more concretely document
        # USB_TYPE_VENDOR: {
        #    USB_REQ_GET_STATUS: "GET_STATUS",
        #    USB_REQ_SET_FEATURE: "SET_FEATURE",
        #    USB_REQ_CLEAR_FEATURE: "CLEAR_FEATURE",
        #    USB_REQ_SYNCH_FRAME: "SYNCH_FRAME",
        #},
    }
    if vendor:
        m[USB_TYPE_VENDOR].update(vendor)

    reqType = bRequestType & USB_TYPE_MASK
    n = m.get(reqType, None)
    if n is None or not bRequest in n:
        return None
    reqs = n[bRequest]
    return reqs


def request_type2str(bRequestType):
    ret = ""

    if (bRequestType & USB_DIR_IN) == USB_DIR_IN:
        ret += "USB_DIR_IN"
    else:
        ret += "USB_DIR_OUT"

    m = {
        USB_TYPE_STANDARD: " | USB_TYPE_STANDARD",
        USB_TYPE_CLASS: " | USB_TYPE_CLASS",
        USB_TYPE_VENDOR: " | USB_TYPE_VENDOR",
        USB_TYPE_RESERVED: " | USB_TYPE_RESERVED",
    }
    ret += m[bRequestType & USB_TYPE_MASK]

    m = {
        USB_RECIP_DEVICE: " | USB_RECIP_DEVICE",
        USB_RECIP_INTERFACE: " | USB_RECIP_INTERFACE",
        USB_RECIP_ENDPOINT: " | USB_RECIP_ENDPOINT",
        USB_RECIP_OTHER: " | USB_RECIP_OTHER",
        USB_RECIP_PORT: " | USB_RECIP_PORT",
        USB_RECIP_RPIPE: " | USB_RECIP_RPIPE",
    }
    ret += m[bRequestType & USB_RECIP_MASK]

    return ret


transfer2str = {
    URB_ISOCHRONOUS: "URB_ISOCHRONOUS",
    URB_INTERRUPT: "URB_INTERRUPT",
    URB_CONTROL: "URB_CONTROL",
    URB_BULK: "URB_BULK",
}


def transfer2str_safe(t):
    return transfer2str.get(t, "UNKNOWN_%02x" % t)
